"""
Aligns nodes into circle
Version 1.0.0: Author: Evan Yip and Jin Xu (2021)

"""
import wx
from rkviewer.plugin.classes import PluginMetadata, WindowedPlugin, PluginCategory
from rkviewer.plugin import api
from rkviewer.plugin.api import Vec2
import numpy as np
import math


class AlignCircle(WindowedPlugin):
    metadata = PluginMetadata(
        name='AlignCircle',
        author='Evan Yip and Jin Xu',
        version='1.0.0',
        short_desc='Align Circle',
        long_desc='Aligns the nodes into a circle',
        category=PluginCategory.UTILITIES
    )

    def __init__(self):
        """
        Initialize the AlignCircle

        Args:
            self
        """
        # allows the align circle class to inherit
        # the methods of the Windowed Plugin Class
        super().__init__()

    def create_window(self, dialog):
        """
        Create a window to do the structural analysis.
        Args:
            self
            dialog
        """
        # Making the window
        window = wx.Panel(dialog, pos=(5, 100), size=(300, 155))
        wx.StaticText(window, -1, 'Select nodes to arrange in circle',
                      (15, 10))

        # Use default radius
        wx.StaticText(window, -1, 'Use default radius:', (15, 30))
        self.defaultCheck = wx.CheckBox(window, -1, pos=(150, 30))
        self.defaultCheck.SetValue(True)

        # Making radius setable
        wx.StaticText(window, -1, 'Input desired radius:', (15, 55))
        self.radiusText = wx.TextCtrl(window, -1, '0', (150, 55),
                                      size=(120, 22))
        self.radiusText.SetInsertionPoint(0)
        self.radiusText.Bind(wx.EVT_TEXT, self.OnText_radius)  # binding test
        self.radiusValue = float(self.radiusText.GetValue())

        # Making the toggle button
        apply_btn = wx.ToggleButton(window, -1, 'Apply', (100, 85),
                                    size=(80, 22))
        apply_btn.SetValue(False)
        # Binding the method to the button
        apply_btn.Bind(wx.EVT_TOGGLEBUTTON, self.Apply)
        return window

    def find_center(self, num_nodes, nodes):
        """
        Takes in the number of nodes and list of node indices and
        computes the optimal centerpoint for the circle
        Parameters: num_nodes(int) - number of nodes,
            node_indices(list) - list of nodes indices
        Returns: center(tuple)
        """
        # Max dimension of the node size
        max_dim = 0  # will consider this the diameter of a circle node
        for i in nodes:
            # Get node
            node = api.get_node_by_index(0, i)
            for dim in node.size:
                if dim > max_dim:
                    max_dim = dim

        # Approximate circumference estimate
        spacing = max_dim / 4
        circum_estimate = num_nodes * (max_dim + spacing)

        # Computing radius
        r = circum_estimate / (2*np.pi) + max_dim

        center = (r, r)
        return center

    def cart(self, r, theta):
        """
        Converts from polar coordinates to cartesian coordinates
        Parameters: r(double), theta(double in radians)
        Returns: (x,y) cartesian coordinate tuple
        """
        x = r * math.cos(theta)
        y = r * math.sin(theta)
        return x, y

    def get_new_position(self, node_index, r, theta):
        """
        Takes in the node index and outputs the new position for that node
        Parameters: r(double), theta(double in radians)
        Returns: (x,y) cartesian coordinate tuple
        """
        node = api.get_node_by_index(0, node_index)
        size = node.size  # (width, height)
        # accounting for slight offset, converting to cartesian
        nodeCenter = self.cart(r - size[0], theta)
        # accounting for node position being specified by top left corner
        x = r + nodeCenter[0] - size[0]/2
        y = r + nodeCenter[1] + size[1]/2
        return Vec2(x, y)

    def OnText_radius(self, event):
        """
        Catches exception if self.radiusText can not be converted
        to a floating point number. Opens a window.
        """
        update = event.GetString()
        if update != '':
            try:
                self.radiusValue = float(self.radiusText.GetValue())
            except:
                wx.MessageBox("Please enter a floating point number"
                              "for the desired radius", "Message",
                              wx.OK | wx.ICON_INFORMATION)

    def CheckSelection(self, nodes):
        """
        Verifies that there are selected nodes. Raises window if
        no nodes are selected, but apply button is pressed
        """
        if nodes == 0:
            wx.MessageBox("Please select desired nodes to arrange"
                          "in circle", "Message", wx.OK | wx.ICON_INFORMATION)
            return True

    def Apply(self, event):
        """
        If apply button is clicked, the nodes will be arranged in a circle
        using either a default radius or a user input radius.
        """
        def translate_nodes(node_indices, r, phi):
            """
            Takes in list of node indices, desired radius, and phi
            and moves the current node
            to its new position
            """
            # Iterate through the nodes and change their position.
            node_num = 0
            for i in node_ind:
                theta = node_num * phi  # angle position for node
                newPos = self.get_new_position(i, r, theta)
                if newPos[0] < 0 or newPos[1] < 0:
                    wx.MessageBox("Please increase radius size",
                                  "Message", wx.OK | wx.ICON_INFORMATION)
                    return
                api.move_node(0, i, newPos, False)
                node_num += 1  # updating node number

        # Get button and checkbox state
        btn_state = event.GetEventObject().GetValue()
        # chk_state = self.OnDefaultCheck()
        chk_state = self.defaultCheck.GetValue()

        # If the button is pressed, arrange the nodes into a circle
        if btn_state is True:
            # Get number of nodes selected
            node_len = len(api.selected_node_indices())
            # get list of node indices
            node_ind = api.selected_node_indices()

            # If no nodes selected raise error
            select = self.CheckSelection(node_len)
            if select is True:
                return

            # If default radius is checked
            if chk_state is True:
                # Compute the expected center of the circle
                center = self.find_center(node_len, node_ind)
                # Compute the angle step between each node
                phi = 2 * math.pi / node_len
                r = center[0]  # radius
                translate_nodes(node_ind, r, phi)

            else:
                r = self.radiusValue
                center = Vec2(r, r)
                phi = 2 * math.pi / node_len  # angle step between each node
                translate_nodes(node_ind, r, phi)  # translating nodes

            # Making the reactions straight lines
            rxns = api.get_reaction_indices(0)
            for rxn in rxns:
                api.update_reaction(net_index=0, reaction_index=rxn,
                                    use_bezier=False)
            # Setting button state back to False (unclicked)
            event.GetEventObject().SetValue(False)



from rkviewer.canvas.elements import CanvasElement as CanvasElement_
from .api import get_canvas
from rkviewer.canvas.utils import draw_rect

CanvasElement = CanvasElement_


def add_element(net_index: int, element: CanvasElement):
    """Add a CanvasElement to the canvas"""
    get_canvas().AddPluginElement(net_index, element)


def remove_element(net_index: int, element: CanvasElement):
    """Remove a CanvasElement from the canvas"""
    get_canvas().RemovePluginElement(net_index, element)



# pylint: disable=maybe-no-member
import wx
import sys
import logging
import logging.config
import traceback
from rkviewer.view import RKView
from rkviewer.controller import Controller
import os
from pathlib import Path

sys.path.append('.')

class ExceptionDialog(wx.MessageDialog):
    def __init__(self, msg):
        """Constructor"""
        super().__init__(None, msg, "Unknown Exception", wx.OK | wx.ICON_ERROR)


# for monkey-patching exceptions
def create_excepthook(old_excepthook):
    dlg = None
    # whether we've already displayed the dialog once, since wx might take some time to die
    over = False

    def custom_excepthook(etype, value, tb):
        nonlocal over, dlg
        if over:
            return
        over = True
        err_msg = ''.join(traceback.format_exception(etype, value, tb))
        logging.error(err_msg)
        #old_excepthook(etype, value, traceback)
        if dlg is None:
            dlg = ExceptionDialog(err_msg)
            dlg.ShowModal()

            wx.GetApp().GetTopWindow().Destroy()

            dlg.Destroy()
            dlg = None

    return custom_excepthook


def setup_logging():
    d = {
        'version': 1,
        'formatters': {
            'detailed': {
                'class': 'logging.Formatter',
                'format': '%(asctime)s %(name)-15s %(levelname)-8s %(message)s'
            },
            'plugins': {
                'class': 'logging.Formatter',
                'format': '%(asctime)s %(filename)-15s %(levelname)-8s %(message)s',
            },
        },
        'handlers': {
            'console': {
                'class': 'logging.StreamHandler',
                'level': 'INFO',
                'stream': 'ext://sys.stdout',
            },
            'debug': {
                'class': 'logging.FileHandler',
                'filename': 'rkviewer.log',
                'mode': 'w',
                'level': 'DEBUG',
                'formatter': 'detailed',
            },
            # 'errors': {
            #     'class': 'logging.FileHandler',
            #     'filename': 'rkviewer-errors.log',
            #     'mode': 'w',
            #     'level': 'ERROR',
            #     'formatter': 'detailed',
            # },
            'plugin-debug': {
                'class': 'logging.FileHandler',
                'filename': 'rkviewer-plugins.log',
                'mode': 'w',
                'level': 'DEBUG',
                'formatter': 'plugins',
            }
        },
        'loggers': {
            'controller': {
                'handlers': ['debug']
            },
            'canvas': {
                'handlers': ['debug']
            },
            'plugin': {
                'handlers': ['plugin-debug']
            },
        },
        'root': {
            'level': 'DEBUG',
            'handlers': ['console', 'debug']
        },
    }
    logging.config.dictConfig(d)


def main():
    setup_logging()

    # global old_excepthook
    # old_excepthook = sys.excepthook
    sys.excepthook = create_excepthook(sys.excepthook)

    logging.info('Initializing SBcoyote...')
    view = RKView()
    controller = Controller(view)
    view.bind_controller(controller)
    view.init()
    view.main_loop()


if __name__ == '__main__':
    main()



# from __future__ import annotations
# pylint: disable=maybe-no-member
from abc import abstractmethod
import enum
from functools import partial
from itertools import chain
from math import pi, cos, sin
from typing import Any, Callable, Dict, Iterable, Iterator, List, Optional, Set, Tuple, Union, cast
from copy import copy

from numpy.lib.utils import source
from rkviewer.canvas.data import CirclePrim, CompositeShape, RectanglePrim, TrianglePrim, Transform, TextPrim, HexagonPrim, LinePrim

import wx

from ..config import Color, get_setting, get_theme
from ..events import (
    CanvasEvent, DidChangeCompartmentOfNodesEvent, DidCommitDragEvent, DidMoveBezierHandleEvent, DidMoveReactionCenterEvent, DidResizeCompartmentsEvent, DidResizeNodesEvent, DidMoveCompartmentsEvent,
    DidMoveNodesEvent, bind_handler,
    post_event, unbind_handler,
)
from ..mvc import IController
from ..utils import change_opacity, even_round, gchain, int_round
from .data import Compartment, HandleData, ModifierTipStyle, Node, Reaction, ReactionBezier, RectData, SpeciesBezier, PolygonPrim, TextAlignment, TextPosition
from .geometry import (
    Rect,
    Vec2,
    clamp_point, clamp_rect_pos,
    get_bounding_rect,
    padded_rect,
    pt_in_circle,
    pt_in_rect, rotate_unit, segment_rect_intersection,
)
from .state import InputMode, cstate
from .utils import draw_rect

# SetCursorFn = Callable[[wx.Cursor], None]
Layer = Union[int, Tuple[int, ...]]


def layer_above(layer: Layer, count: int = 1) -> Layer:
    """
    Return the next layer above this layer, without increasing the length of the layer list.

    count is optionally the layer number increment.
    """
    if count < 1:
        raise ValueError('Layer count must be at least 1!')
    if isinstance(layer, int):
        return layer + count
    else:
        last = len(layer) - 1
        return layer[0:last] + (layer[last] + count,)


class CanvasElement:
    """Base class for an element positioned on the canvas.

    Attributes:
        layers: The layer(s) number of this element.
        enabled: Whether the element is enabled.
        destroyed: Whether the object was destroyed (if this is True then you shouldn't use this)
    """
    layers: Layer
    enabled: bool
    destroyed: bool

    def __init__(self, layers: Layer):
        if isinstance(layers, int):
            layers = (layers,)
        self.layers = layers
        self.enabled = True
        self.destroyed = False

    def set_layers(self, layers: Layer):
        if isinstance(layers, int):
            layers = (layers,)
        self.layers = layers

    def destroy(self):
        """Destroy this element; override this for specific implementations."""
        self.destroyed = True

    def pos_inside(self, logical_pos: Vec2) -> bool:
        """Returns whether logical_pos is inside the diplayed shape of this element."""
        return False

    @abstractmethod
    def on_paint(self, gc: wx.GraphicsContext):
        """Paint the shape onto the given GraphicsContext.

        This draws onto the scrolled canvas, i.e. the position of the drawn item will respond to
        scrolling, so you don't need to account for that.
        """
        pass

    def on_paint_cue(self, gc: wx.GraphicsContext):
        '''This is called to paint special visual cues tha tthe element may need to display'''
        pass

    def on_mouse_enter(self, logical_pos: Vec2) -> bool:
        """Handler for when the mouse has entered the shape."""
        return False

    def on_mouse_leave(self, logical_pos: Vec2) -> bool:
        """Handler for when the mouse has exited the shape"""
        return False

    def on_mouse_move(self, logical_pos: Vec2) -> bool:
        """Handler for when the mouse moves inside the shape, with the left mouse button up."""
        return False

    def on_mouse_drag(self, logical_pos: Vec2, rel_pos: Vec2) -> bool:
        """Handler for when the mouse drags inside the shape, with the left mouse button down."""
        return False

    def on_left_down(self, logical_pos: Vec2) -> bool:
        """Handler for when the mouse left button is pressed down inside the shape."""
        return False

    def on_left_up(self, logical_pos: Vec2) -> bool:
        """Handler for when the mouse left button is springs up inside the shape."""
        return False

    def bounding_rect(self) -> Rect:
        """Return the bounding rectangle of the element."""
        return Rect(Vec2(), Vec2())


class NodeElement(CanvasElement):
    """CanvasElement for nodes."""
    node: Node
    canvas: Any

    # HACK no type specified for canvas since otherwise there would be circular dependency
    def __init__(self, node: Node, canvas, layers: Layer):
        super().__init__(layers)
        self.node = node
        self.canvas = canvas
        self.gfont = None  # In the future
        self.font_scale = 1

    def pos_inside(self, logical_pos: Vec2) -> bool:
        return pt_in_rect(logical_pos, self.node.s_rect)

    def on_paint(self, gc: wx.GraphicsContext):
        if self.gfont is None or self.font_scale != cstate.scale:
            font = wx.Font(wx.FontInfo(10))
            self.gfont = gc.CreateFont(font, wx.BLACK)
        gc.SetFont(self.gfont)

        # boundaryFactor = 1
        # if not self.node.floatingNode:
        #     boundaryFactor = 2  # Store this in a theme?

        s_aligned_rect = self.node.s_rect.aligned()
        # aligned_border_width = max(even_round(
        #     self.node.border_width * boundaryFactor), 2)
        width, height = s_aligned_rect.size

        assert self.node.composite_shape is not None
        draw_composite_shape(
            gc,
            self.node.rect,
            self.node)

        if self.node.lockNode:
            lock_color = self.node.border_color or Color(255, 0, 0)
            pen = gc.CreatePen(wx.GraphicsPenInfo(
                lock_color.to_wxcolour()).Width(2))
            gc.SetPen(pen)
            path = gc.CreatePath()
            path.AddCircle(self.node.position.x, self.node.position.y, .1*height)
            gc.StrokePath(path)

        if not self.node.floatingNode:
            boundary_color = self.node.border_color or Color(255, 0, 0)
            pen = gc.CreatePen(wx.GraphicsPenInfo(
                boundary_color.to_wxcolour()).Width(2))
            gc.SetPen(pen)
            path = gc.CreatePath()
            path.AddRectangle(self.node.position.x-round(self.node.border_width/2 + 2)-2,
                              self.node.position.y-round(self.node.border_width/2 + 2)-2,
                              2*round(self.node.border_width/2 + 2)+self.node.size.x+4,
                              2*round(self.node.border_width/2 + 2)+self.node.size.y+4)
            gc.StrokePath(path)

    def on_left_down(self, _: Vec2):
        return True

    def bounding_rect(self) -> Rect:
        return self.node.rect


class BezierHandle(CanvasElement):
    """Class that keeps track of a Bezier control handle tip.

    Attributes:
        HANDLE_RADIUS: radius of the control handle.
        data: The associated HandleData>
        on_moved: The function called when the handle is moved.
        on_dropped: The function called when the handle is dropped (i.e. mouse up).
        reaction: The associated Reaction.
        twin: The twin BezierHandle; used only for the center handles.
        node_idx: The index of the node associated with this handle. -1 if this is a source
                  centroid handle, and -2 if this is a target centroid handle.
    """
    HANDLE_RADIUS = 5  # Radius of the control handle

    data: HandleData
    on_moved: Callable[[Vec2], None]
    on_dropped: Callable[[Vec2], None]
    reaction: Reaction
    twin = Any
    node_idx: int

    def __init__(self, data: HandleData, layer: Layer, on_moved: Callable[[Vec2], None],
                 on_dropped: Callable[[Vec2], None], reaction: Reaction, node_idx: int):
        super().__init__(layer)
        self.data = data
        self.on_moved = on_moved
        self.on_dropped = on_dropped
        self.reaction = reaction
        self.hovering = False
        self.enabled = False
        self.twin = None
        self.node_idx = node_idx

    def pos_inside(self, logical_pos: Vec2):
        return pt_in_circle(logical_pos, BezierHandle.HANDLE_RADIUS, self.data.tip)

    def _paint(self, gc: wx.GraphicsContext, handle_color: wx.Colour):
        brush = wx.Brush(handle_color)
        pen = gc.CreatePen(wx.GraphicsPenInfo(handle_color))

        sbase = self.data.base
        stip = self.data.tip

        gc.SetPen(pen)

        # Draw handle lines
        gc.StrokeLine(*sbase, *stip)

        # Draw handle circles
        gc.SetBrush(brush)
        gc.DrawEllipse(stip.x - BezierHandle.HANDLE_RADIUS,
                        stip.y - BezierHandle.HANDLE_RADIUS,
                        2 * BezierHandle.HANDLE_RADIUS, 2 * BezierHandle.HANDLE_RADIUS)

    def on_paint(self, gc: wx.GraphicsContext):
        """Paint the handle as given by its base and tip positions, highlighting it if hovering."""
        assert self.data.base is not None
        if self.reaction.bezierCurves:
            c = get_theme('handle_color')
            self._paint(gc, c)

    def on_paint_cue(self, gc: wx.GraphicsContext):
        if self.hovering and self.reaction.bezierCurves:
            assert self.data.base is not None
            c = get_theme('highlighted_handle_color')
            self._paint(gc, c)

    def on_mouse_enter(self, logical_pos: Vec2) -> bool:
        self.hovering = True
        if self.twin:
            self.twin.hovering = True
        return True

    def on_mouse_leave(self, logical_pos: Vec2) -> bool:
        self.hovering = False
        if self.twin:
            self.twin.hovering = False
        return True

    def on_left_down(self, logical_pos: Vec2) -> bool:
        return True

    def on_mouse_drag(self, logical_pos: Vec2, rel_pos: Vec2) -> bool:
        self.data.tip += rel_pos
        self.on_moved(self.data.tip)
        neti = 0
        post_event(DidMoveBezierHandleEvent(neti, self.reaction.index,
                                            self.node_idx, by_user=True, direct=True))

        return True

    def on_left_up(self, logical_pos: Vec2):
        self.on_dropped(self.data.tip)
        return True


class ReactionCenter(CanvasElement):
    parent: 'ReactionElement'
    _moved: bool

    def __init__(self, parent: 'ReactionElement', layers: Layer):
        super().__init__(layers)
        self.parent = parent
        self._moved = False
        self.hovering = False

    def _paint(self, gc: wx.GraphicsContext, color: wx.Colour):
        pen = wx.Pen(color)
        brush = wx.Brush(color)
        gc.SetPen(pen)
        gc.SetBrush(brush)
        radius = get_theme('reaction_radius')
        center = self.parent.bezier.real_center - Vec2.repeat(radius)
        gc.DrawEllipse(center.x, center.y, radius * 2, radius * 2)

    def on_paint(self, gc: wx.GraphicsContext):
        if not self.parent.selected:
            return

        # draw centroid
        color = self.parent.reaction.fill_color
        if self.parent.selected:
            color = get_theme('handle_color')
        self._paint(gc, color)

    def on_paint_cue(self, gc: wx.GraphicsContext):
        if self.parent.selected and self.hovering:
            color = get_theme('highlighted_handle_color')
            self._paint(gc, color)

    def on_left_down(self, logical_pos: Vec2) -> bool:
        # If not selected, then nothing is done to prevent accidental dragging
        return True

    def on_mouse_drag(self, logical_pos: Vec2, rel_pos: Vec2) -> bool:
        offset = rel_pos
        reaction = self.parent.reaction
        reaction.center_pos = self.parent.bezier.real_center + offset
        self.parent.bezier.center_moved(offset)
        self._moved = True
        net_index = 0
        post_event(DidMoveReactionCenterEvent(net_index, reaction.index, offset, True))
        return True

    def on_left_up(self, logical_pos: Vec2) -> bool:
        ctrl = self.parent.controller
        neti = self.parent.canvas.net_index
        reai = self.parent.reaction.index
        with ctrl.group_action():
            ctrl.set_reaction_center(neti, reai, self.parent.reaction.center_pos)
            ctrl.set_center_handle(neti, reai, self.parent.reaction.src_c_handle.tip)
            post_event(DidCommitDragEvent(self))
        self._moved = False
        return True

    def pos_inside(self, logical_pos: Vec2) -> bool:
        # TODO works witih zoom?
        radius = get_theme('reaction_radius')
        return pt_in_circle(self.parent.bezier.real_center, radius, logical_pos)

    def on_mouse_enter(self, logical_pos: Vec2) -> bool:
        self.hovering = True
        return True

    def on_mouse_leave(self, logical_pos: Vec2) -> bool:
        self.hovering = False
        return True

    def bounding_rect(self) -> Rect:
        radius = get_theme('reaction_radius')
        return Rect(self.parent.bezier.real_center - radius, Vec2.repeat(radius) * 2)


# Uniquely identifies nodes in a reaction: (regular node index; whether the node is a reactant)
RIndex = Tuple[int, bool]


class ReactionElement(CanvasElement):
    """CanvasElement for reactions.

    Note that if new nodes are constructed, all ReactionBezier instances that use these nodes should
    be re-constructed with the new nodes. On the other hand, if the nodes are merely modified, the
    corresponding update methods should be called.
    """
    reaction: Reaction
    center_el: ReactionCenter
    index_to_bz: Dict[RIndex, SpeciesBezier]
    bezier: ReactionBezier
    bhandles: List[BezierHandle]
    moved_handler_id: int
    #: Set of indices of the nodes that have been moved, but not committed.
    _dirty_node_indices: Set[int]
    #: Works in tandem with _dirty_indices. True if all nodes of the reaction are being moved.
    _moving_all: bool
    _selected: bool
    canvas: Any  # avoid circular dependency
    controller: IController

    # HACK no type for canvas since otherwise there is circular dependency
    def __init__(self, reaction: Reaction, bezier: ReactionBezier, canvas, layers: Layer,
                 handle_layer: Layer):
        super().__init__(layers)
        self.reaction = reaction
        self.containing_node_indices = set(chain(reaction.sources, reaction.targets))
        self.bezier = bezier
        self.moved_handler_id = bind_handler(DidMoveNodesEvent, self.nodes_moved)
        # i is 0 for source Beziers, but 1 for dest Beziers. "not" it to get the correct bool.
        self.index_to_bz = {(bz.node_idx, not gi): bz
                            for gi, bz in gchain(bezier.src_beziers,
                                                 bezier.dest_beziers)}
        self.canvas = canvas
        self.controller = canvas.controller
        self._hovered_handle = None
        self._dirty_indices = set()
        self._moving_all = False
        self.bhandles = list()
        self._selected = False

        neti = canvas.net_index
        reai = reaction.index
        ctrl = canvas.controller
        # create elements for species
        for gi, sb in gchain(bezier.src_beziers, bezier.dest_beziers):
            dropped_func = self.make_drop_handle_func(ctrl, neti, reai, sb.node_idx, not gi)
            el = BezierHandle(sb.handle, handle_layer,
                              bezier.make_handle_moved_func(sb), dropped_func, reaction, sb.node_idx)
            self.bhandles.append(el)

        def centroid_handle_dropped(p: Vec2):
            with ctrl.group_action():
                ctrl.set_center_handle(neti, reai, reaction.src_c_handle.tip)
                # NOTE important to do this within the group action
                post_event(DidCommitDragEvent(self))

        src_bh = BezierHandle(reaction.src_c_handle, handle_layer,
                              lambda _: bezier.src_handle_moved(),
                              centroid_handle_dropped, reaction, -1)
        dest_bh = BezierHandle(reaction.dest_c_handle, handle_layer,
                               lambda _: bezier.dest_handle_moved(),
                               centroid_handle_dropped, reaction, -2)
        src_bh.twin = dest_bh
        dest_bh.twin = src_bh
        self.bhandles.append(src_bh)
        self.bhandles.append(dest_bh)
        center_layers = layer_above(handle_layer, count=2)
        self.center_el = ReactionCenter(self, center_layers)

    def make_drop_handle_func(self, ctrl: IController, neti: int, reai: int, nodei: int,
                              is_source: bool):
        if is_source:
            def ret(p):
                ctrl.set_src_node_handle(neti, reai, nodei, p)
                post_event(DidCommitDragEvent(self))
            return ret
        else:
            # Avoid IDE warnings
            def ret1(p):
                ctrl.set_dest_node_handle(neti, reai, nodei, p)
                post_event(DidCommitDragEvent(self))
            return ret1

    @property
    def selected(self) -> bool:
        return self._selected

    @selected.setter
    def selected(self, val: bool):
        # Enable/disable Handles based on whether the curve is selected
        self._selected = val
        for bz in self.bhandles:
            bz.enabled = val
        self.center_el.enabled = val

    def nodes_moved(self, evt: CanvasEvent):
        """Handler for after a node has moved."""
        # If already moving (i.e. self._dirty_indices is not empty), then skip forward
        c_evt = cast(DidMoveNodesEvent, evt)
        node_indices = set(c_evt.node_indices) & self.containing_node_indices
        if not node_indices:
            return
        offset = c_evt.offset
        rects = [self.canvas.node_idx_map[idx].rect for idx in chain(
            self.reaction.sources, self.reaction.targets)]
        self.bezier.nodes_moved(rects)
        if len(self._dirty_indices) == 0:
            self._dirty_indices = node_indices
            my_indices = set(chain(self.reaction.sources, self.reaction.targets))
            self._moving_all = my_indices <= self._dirty_indices

        neti = 0
        for i, idx in enumerate(node_indices):
            for in_src in [True, False]:
                if (idx, in_src) in self.index_to_bz:
                    bz = self.index_to_bz[(idx, in_src)]
                    off = offset if isinstance(offset, Vec2) else offset[i]
                    bz.handle.tip += off
                    post_event(DidMoveBezierHandleEvent(neti, self.reaction.index, bz.node_idx,
                                                        by_user=True, direct=False))
                    bz.update_curve(self.bezier.real_center)

        if self._moving_all and isinstance(offset, Vec2):
            # Only move src_handle_tip if moving all nodes and they are moved by the same amount.
            self.reaction.src_c_handle.tip += offset
            # move center pos as well if it is not auto-set as the centroid
            if self.reaction.center_pos:
                self.reaction.center_pos += offset
            self.bezier.src_handle_moved()
            post_event(DidMoveBezierHandleEvent(neti, self.reaction.index, -1, by_user=True,
                                                direct=False))

    def commit_node_pos(self):
        """Handler for after the controller is told to move a node."""
        ctrl: IController = self.canvas.controller
        neti = self.canvas.net_index
        reai = self.reaction.index
        for bz in self.bezier.src_beziers:
            if bz.node_idx in self._dirty_indices:
                ctrl.set_src_node_handle(neti, reai, bz.node_idx, bz.handle.tip)

        for bz in self.bezier.dest_beziers:
            if bz.node_idx in self._dirty_indices:
                ctrl.set_dest_node_handle(neti, reai, bz.node_idx, bz.handle.tip)

        if self._moving_all:
            ctrl.set_center_handle(neti, reai, self.reaction.src_c_handle.tip)
            if self.reaction.center_pos:
                ctrl.set_reaction_center(neti, reai, self.reaction.center_pos)
        self._dirty_indices = set()

    def destroy(self):
        unbind_handler(self.moved_handler_id)
        super().destroy()

    def pos_inside(self, logical_pos: Vec2) -> bool:
        return self.bezier.is_mouse_on(logical_pos)

    def on_mouse_enter(self, logical_pos: Vec2):
        self.on_mouse_move(logical_pos)

    def on_left_down(self, logical_pos: Vec2) -> bool:
        return True  # Return True so that this can be selected

    def on_paint(self, gc: wx.GraphicsContext):
        self.bezier.do_paint(gc, self.reaction.fill_color, self.selected)

        MOD_NODE_PAD = 10
        RXN_PAD = 15
        MODIFIER_RADIUS = 3
        TEE_LENGTH = 5

        line_width = get_theme('modifier_line_width')
        pen = gc.CreatePen(wx.GraphicsPenInfo(get_theme('modifier_line_color')).Width(line_width))
        brush = gc.CreateBrush(wx.Brush(get_theme('modifier_line_color')))
        gc.SetPen(pen)
        gc.SetBrush(brush)
        # Draw modifier lines
        for modifier in self.reaction.modifiers:
            mod_node = self.canvas.node_idx_map[modifier]
            rect = mod_node.rect
            clipping_rect = padded_rect(rect, MOD_NODE_PAD)

            node_center = rect.center_point
            rxn_center = self.bezier.real_center

            # If too close, don't draw. Pad rectangle and circle with 1 to avoid floating point
            # shenanigans
            if pt_in_rect(rxn_center, padded_rect(clipping_rect, 1)):
                continue
            elif pt_in_circle(node_center, RXN_PAD + 1, rxn_center):
                continue

            # create segment
            segment = (node_center, rxn_center)
            diff = node_center - rxn_center
            rxn_intersection = (rxn_center + diff.normalized(RXN_PAD))
            node_intersection = segment_rect_intersection(segment, clipping_rect)

            # gc.StrokeLine(*node_intersection, *rxn_intersection)
            path = gc.CreatePath()
            path.MoveToPoint(*node_intersection)
            path.AddLineToPoint(*rxn_intersection)
            if self.reaction.modifier_tip_style == ModifierTipStyle.CIRCLE:
                path.AddCircle(*rxn_intersection, MODIFIER_RADIUS)
            elif self.reaction.modifier_tip_style == ModifierTipStyle.TEE:
                # draw T-shaped tip
                ortho = rotate_unit(diff, pi / 2)
                pt1 = rxn_intersection + ortho * TEE_LENGTH
                pt2 = rxn_intersection - ortho * TEE_LENGTH
                path.MoveToPoint(*pt1)
                path.AddLineToPoint(*pt2)
            gc.StrokePath(path)
            gc.FillPath(path)
            # path.MoveToPoint(0.0, 50.0)

    def bounding_rect(self) -> Rect:
        return self.bezier.get_bounding_rect()


class CompartmentElt(CanvasElement):
    def __init__(self, compartment: Compartment, major_layer: int, minor_layer: int):
        super().__init__((major_layer, minor_layer))
        self.compartment = compartment

    def pos_inside(self, logical_pos: Vec2) -> bool:
        return pt_in_rect(logical_pos, self.compartment.rect)

    def on_left_down(self, logical_pos: Vec2) -> bool:
        return True

    def _paint(self, gc: wx.GraphicsContext, highlight: bool):
        rect = Rect(self.compartment.position,
                    self.compartment.size)
        if highlight:
            border = wx.Colour(255, 255, 255, 100)
            fill = wx.Colour(255, 255, 255, 100)
        else:
            border = self.compartment.border
            fill = self.compartment.fill
        draw_rect(gc, rect, border=border,
                  border_width=self.compartment.border_width, fill=fill,
                  corner_radius=get_theme('comp_corner_radius'))

    def on_paint(self, gc: wx.GraphicsContext):
        self._paint(gc, False)

    def highlight_paint(self, gc: wx.GraphicsContext):
        self._paint(gc, True)

    def bounding_rect(self) -> Rect:
        return self.compartment.rect


class SelectBox(CanvasElement):
    """Class that represents a select box, i.e. the bounding box draw around the selected nodes.

    Supports moving and resizing operations.

    Attributes:
        CURSOR_TYPES: List of cursor types starting with that for the top-left handle and going
                      clockwise.
        nodes: List of selected nodes, as contained in this select box.
        related_elts: List of NodeElements related to each node instance; matches the node list
                      1-1.
        bounding_rect: The exact bounding rectangle (without padding).
        mode: Current input mode of the SelectBox.

    Note:
        The behavior of the SelectBox depends on the nodes and compartments selected, but not the
        reactions. The cases of behaviors are documented here:

        1) Only compartments are selected. Nodes within these compartments are dragged along with
           them, but they are not resized.
        3) Only nodes are selected, and they are all in the same compartment. In this case, the
           nodes may be moved normally. They also may be moved outside of their compartment to be
           assigned to another compartment (this is the only case where this is possible). However
           note that the nodes may not be resized to be larger than the containing compartment.
        3) Otherwise, there are two cases depending on if the selected nodes are in the union of
           the selected compartments.
            a) If the selected nodes are entirely contained in the list of selected compartments,
               then everything is moved and resized together, as usual.
            b) Otherwise (i.e. some node is not in any selected compartment), then dragging and
               resizing are disabled.
        Note that in case 2), if all selected nodes are in the base compartment (i.e. no
        compartment), then the base compartment is assumed to be selected, and resizing and moving
        work as usual.
    """
    CURSOR_TYPES = [wx.CURSOR_SIZENWSE, wx.CURSOR_SIZENS, wx.CURSOR_SIZENESW, wx.CURSOR_SIZEWE,
                    wx.CURSOR_SIZENWSE, wx.CURSOR_SIZENS, wx.CURSOR_SIZENESW, wx.CURSOR_SIZEWE]
    HANDLE_MULT = [Vec2(), Vec2(1/2, 0), Vec2(1, 0), Vec2(1, 1/2),
                   Vec2(1, 1), Vec2(1/2, 1), Vec2(0, 1), Vec2(0, 1/2)]

    nodes: List[Node]
    node_indices: List[int]
    compartments: List[Compartment]
    comp_indices: List[int]
    # List of nodes that are not selected, but are within selected compartments. Used only
    # for SMode.CONTAINED
    peripheral_nodes: List[Node]
    related_elts: List[CanvasElement]
    bounding_rect: Rect
    _padding: float  #: padding for the bounding rectangle around the selected nodes
    _drag_rel: Vec2  #: relative position of the mouse to the bounding rect when dragging started
    #: whether the node was drag-moved between left_down and left_up.
    _did_move: bool
    _orig_rpos: List[Vec2]  #: relative positions of the comps and nodes to the select box
    _orig_rsize: List[Vec2]  #: sizes of the comps and nodes to the select box
    #: relative positions of the compartments and nodes to cursor; updated when dragging starts
    _rel_positions: List[Vec2]
    _resize_handle: int  #: the node resize handle.
    node_min_ratio: Optional[Vec2]
    comp_min_ratio: Optional[Vec2]
    #: the minimum resize ratio for each axis, to avoid making the nodes too small
    _min_resize_ratio: Vec2
    #: the bounding rect when dragging/resizing started
    _orig_rect: Optional[Rect]
    _bounds: Rect  #: the bounds that the bounding rect may not exceed
    special_mode: 'SelectBox.SMode'

    class Mode(enum.Enum):
        IDLE = 0
        MOVING = 1
        RESIZING = 2

    class SMode(enum.Enum):
        """For what this does, see "Notes" section of the class documentation."""

        COMP_ONLY = 0
        """Only compartments are selected."""
        NODES_IN_ONE = 1
        """Only nodes are selected, and they are in a single compartment."""
        CONTAINED = 2
        """Nodes are entirely contained in the selected compartments, or they are all in the base
        compartment.
        """
        NOT_CONTAINED = 3
        """Nodes are not entirely contained in the selected compartments."""

    def __init__(self, canvas, nodes: List[Node], compartments: List[Compartment], bounds: Rect,
                 controller: IController, net_index: int, layer: int):
        super().__init__(layer)
        self.canvas = canvas
        self.peripheral_nodes = list()
        self.update(nodes, compartments)
        self.controller = controller
        self.net_index = net_index
        self._mode = SelectBox.Mode.IDLE

        self._drag_rel = Vec2()
        self._did_move = False
        self._orig_rpos = list()
        self._orig_rsizes = list()
        self._rel_positions = list()
        self._orig_rect = None
        self._resize_handle = -1
        self._min_resize_ratio = Vec2()
        self._hovered_part = -2

        self._bounds = bounds

    @property
    def mode(self):
        return self._mode

    def update(self, nodes: List[Node], compartments: List[Compartment]):
        self.nodes = nodes
        self.compartments = compartments
        self.node_indices = [n.index for n in self.nodes]
        self.comp_indices = [c.index for c in self.compartments]

        if len(nodes) + len(compartments) > 0:
            if len(nodes) == 1 and len(compartments) == 0:
                # If only one node is selected, reduce padding
                self._padding = get_theme('select_outline_padding')
            else:
                self._padding = get_theme('select_box_padding')
            # Align bounding box if only one node is selected, see NodeElement::on_paint for
            # explanations. Note that self._padding should be an integer
            self._padding = int_round(self._padding)
            self.bounding_rect = get_bounding_rect(
                [n.rect for n in nodes] + [c.rect for c in compartments])

        selected_comps = set(c.index for c in compartments) | {-1}
        # Determine SMode
        if len(nodes) == 0:
            self.special_mode = SelectBox.SMode.COMP_ONLY
        else:
            assoc_comps = {n.comp_idx for n in nodes}
            if len(compartments) == 0:
                if len(assoc_comps) == 1:
                    # Nodes are in one compartment
                    self.special_mode = SelectBox.SMode.NODES_IN_ONE
                else:
                    # Cannot possibly contain
                    self.special_mode = SelectBox.SMode.NOT_CONTAINED
            else:
                if selected_comps >= assoc_comps:
                    self.special_mode = SelectBox.SMode.CONTAINED
                else:
                    self.special_mode = SelectBox.SMode.NOT_CONTAINED

        # Compute peripheral nodes
        if self.special_mode == SelectBox.SMode.NOT_CONTAINED:
            self.peripheral_nodes = list()
        else:
            peri_indices = set()
            for comp in compartments:
                if comp.index in selected_comps:
                    peri_indices |= set(comp.nodes)

            peri_indices -= set(n.index for n in nodes)
            self.peripheral_nodes = [self.canvas.node_idx_map[i] for i in peri_indices]

    def outline_rect(self) -> Rect:
        """Helper that returns the scaled, padded bounding rectangle."""
        return padded_rect((self.bounding_rect).aligned(), self._padding)

    def _resize_handle_rects(self):
        """Helper that computes the scaled positions and sizes of the resize handles.

        Returns:
            A list of (pos, size) tuples representing the resize handle
            rectangles. They are ordered such that the top-left handle is the first element, and
            all other handles follow in clockwise fashion.
        """
        outline_rect = self.outline_rect()
        pos, size = outline_rect.as_tuple()
        centers = [pos, pos + Vec2(size.x / 2, 0),
                   pos + Vec2(size.x, 0), pos + Vec2(size.x, size.y / 2),
                   pos + size, pos + Vec2(size.x / 2, size.y),
                   pos + Vec2(0, size.y), pos + Vec2(0, size.y / 2)]
        centers = [pos + size.elem_mul(m) for m in SelectBox.HANDLE_MULT]
        side = get_theme('select_handle_length')
        return [Rect(c - Vec2.repeat(side/2), Vec2.repeat(side)) for c in centers]

    def _resize_handle_pos(self, n: int):
        pos, size = self.outline_rect().as_tuple()
        assert n >= 0 and n < 8
        return pos + size.elem_mul(SelectBox.HANDLE_MULT[n])

    def _pos_inside_part(self, logical_pos: Vec2) -> int:
        """Helper for determining if logical_pos is within which, if any, part of this widget.

        Returns:
            The handle index (0-3) if pos is within a handle, -1 if pos is not within a handle
            but within the bounding rectangle, or -2 if pos is outside.
        """
        if len(self.nodes) + len(self.compartments) == 0:
            return -2

        rects = self._resize_handle_rects()
        for i, rect in enumerate(rects):
            if pt_in_rect(logical_pos, rect):
                return i

        rects = [n.rect for n in self.nodes] + \
            [c.rect for c in self.compartments]
        if any(pt_in_rect(logical_pos, r) for r in rects):
            return -1
        else:
            return -2

    def pos_inside(self, logical_pos: Vec2):
        return self._pos_inside_part(logical_pos) != -2

    def on_mouse_enter(self, logical_pos: Vec2):
        self.on_mouse_move(logical_pos)

    def on_mouse_move(self, logical_pos: Vec2):
        if cstate.input_mode != InputMode.SELECT:
            return False  # clearly, don't change the cursor if we're not in select mode
        self._hovered_part = self._pos_inside_part(logical_pos)
        if self._hovered_part >= 0:
            cursor = SelectBox.CURSOR_TYPES[self._hovered_part]
            self.canvas.SetCursor(wx.Cursor(cursor))
        elif self._hovered_part == -1:
            self.canvas.SetCursor(wx.Cursor(wx.CURSOR_SIZING))
        return True

    def on_mouse_leave(self, logical_pos: Vec2):
        self._hovered_part = -2
        # HACK re-set input_mode with the same value to make canvas update the cursor
        # See issue #9 for details
        cstate.input_mode = cstate.input_mode
        return True

    def on_paint(self, gc: wx.GraphicsContext):
        if len(self.nodes) + len(self.compartments) > 0:
            outline_width = max(even_round(get_theme('select_outline_width')), 2)
            pos, size = self.outline_rect().as_tuple()

            # draw main outline
            draw_rect(gc, Rect(pos, size), border=get_theme('handle_color'),
                      border_width=outline_width, corner_radius=0)

            for handle_rect in self._resize_handle_rects():
                # convert to device position for drawing
                draw_rect(gc, handle_rect, fill=get_theme('handle_color'))

    def map_rel_pos(self, positions: Iterable[Vec2]) -> List[Vec2]:
        temp = [p - self._orig_rect.position - Vec2.repeat(self._padding)
                for p in positions]
        return temp

    def on_left_down(self, logical_pos: Vec2):
        if len(self.nodes) + len(self.compartments) == 0:
            return False

        # if multi-selecting and clicked on a node/reaction, then the user must mean to de-select
        # that element. In this exceptional case we return False so that canvas can continue the
        # pos_inside loop and find the node/reaction in question later.
        if cstate.multi_select:
            for elt in self.related_elts:
                if elt.pos_inside(logical_pos):
                    return False
        elif len(self.compartments) != 0:
            # make sure that user can select items inside compartment, even if the compartment is
            # itself selected.
            # for elt in self.related_elts:
            #     if elt.pos_inside(logical_pos):
            #         if isinstance(elt, ReactionElement):
            #             return False
            #         elif isinstance(elt, NodeElement) and elt.node.comp_idx != -1:
            #             return False
            #         else:
            #             break
            pass

        handle = self._hovered_part
        # assert self._mode == SelectBox.Mode.IDLE
        if handle >= 0:
            self._mode = SelectBox.Mode.RESIZING
            self._resize_handle = handle
            # calculate minimum resize ratio, enforcing min size constraints on nodes and comps
            self._update_min_resize_ratio()

            #self._orig_rect = self.outline_rect()
            # Take unaligned bounding rect as orig_rect for better accuracy
            self._orig_rect = padded_rect(
                (self.bounding_rect), self._padding)
            # relative starting positions to the select box
            orig_node_pos = self.map_rel_pos((n.position for n in self.nodes))
            orig_comp_pos = self.map_rel_pos((c.position for c in self.compartments))
            orig_node_sizes = [n.size for n in self.nodes]
            orig_comp_sizes = [c.size for c in self.compartments]
            self._orig_rpos = orig_comp_pos + orig_node_pos
            self._orig_rsizes = orig_comp_sizes + orig_node_sizes
            self._resize_handle_offset = self._resize_handle_pos(handle) - logical_pos
            return True
        elif handle == -1:
            self._mode = SelectBox.Mode.MOVING
            # relative starting positions to the mouse positions
            rel_node_pos = [n.position -
                            logical_pos for n in self.nodes if not n.lockNode]
            rel_comp_pos = [c.position - logical_pos for c in self.compartments]
            self._rel_positions = rel_comp_pos + rel_node_pos
            self._drag_rel = self.bounding_rect.position - logical_pos
            return True

        return False

    def compute_min_ratio(self) -> Tuple[Optional[Vec2], Optional[Vec2]]:
        """Compute minimum size ratio resizing nodes and compartments.

        Returns:
            A tuple containing (node mininum resize ratio, comp minimum resize ratio). Each ratio
            may be None, in the case that no elements of that type is selected.
        """
        node_min_ratio = None
        comp_min_ratio = None
        if len(self.nodes) != 0:
            min_width = min(n.size.x for n in self.nodes)
            min_height = min(n.size.y for n in self.nodes)
            node_min_ratio = Vec2(get_setting('min_node_width') / min_width,
                                  get_setting('min_node_height') / min_height)

        if len(self.compartments) != 0:
            min_comp_width = min(c.size.x for c in self.compartments)
            min_comp_height = min(c.size.y for c in self.compartments)
            comp_min_ratio = Vec2(get_setting('min_comp_width') / min_comp_width,
                                  get_setting('min_comp_height') / min_comp_height)
            for node in self.peripheral_nodes:
                comp = self.canvas.comp_idx_map[node.comp_idx]
                ratio = node.size.elem_div(comp.size)
                comp_min_ratio = comp_min_ratio.reduce2(max, ratio)

        return node_min_ratio, comp_min_ratio

    def _update_min_resize_ratio(self):
        node_min_ratio, comp_min_ratio = self.compute_min_ratio()

        if node_min_ratio is not None and comp_min_ratio is not None:
            self._min_resize_ratio = node_min_ratio.reduce2(max, comp_min_ratio)
        elif node_min_ratio is not None:
            self._min_resize_ratio = node_min_ratio
        elif comp_min_ratio is not None:
            self._min_resize_ratio = comp_min_ratio
        else:
            assert False, 'Cannot possibly click on handle when nothing is selected.'

    def on_left_up(self, logical_pos: Vec2):
        assert len(self.nodes) + len(self.compartments) != 0
        if self._mode == SelectBox.Mode.MOVING:
            if self._did_move:
                self._did_move = False

                self._commit_move()
        elif self._mode == SelectBox.Mode.RESIZING:
            assert not self._did_move
            with self.controller.group_action():
                for node in self.peripheral_nodes:
                    self.controller.move_node(self.net_index, node.index, node.position)
                for node in self.nodes:
                    self.controller.move_node(self.net_index, node.index, node.position)
                    self.controller.set_node_size(self.net_index, node.index, node.size)
                for comp in self.compartments:
                    self.controller.move_compartment(self.net_index, comp.index, comp.position)
                    self.controller.set_compartment_size(self.net_index, comp.index, comp.size)
                post_event(DidCommitDragEvent(self))

            # Need to do this in case _special_mode == NOT_CONTAINED, so that the mouse has now
            # left the handle. on_mouse_leave is not triggered because it's considered to be dragging.
            self._hovered_part = self._pos_inside_part(logical_pos)
        self._mode = SelectBox.Mode.IDLE

    def on_mouse_drag(self, logical_pos: Vec2, rel_pos: Vec2) -> bool:
        assert self._mode != SelectBox.Mode.IDLE
        rect_data = cast(List[RectData], self.compartments) + cast(List[RectData], self.nodes)
        if self.special_mode == SelectBox.SMode.NOT_CONTAINED:
            # Return True since we still want this to appear to be dragging, just not working.
            return True
        if self._mode == SelectBox.Mode.RESIZING:
            rect_data = cast(List[RectData], self.compartments) + cast(List[RectData], self.nodes)
            # TODO move the orig_rpos, etc. code to update()
            bounds = self._bounds
            if self.special_mode == SelectBox.SMode.NODES_IN_ONE:
                # constrain resizing to within this compartment
                if self.nodes[0].comp_idx != -1:
                    containing_comp = self.canvas.GetCompartment(self.nodes[0].comp_idx)
                    assert containing_comp is not None
                    bounds = containing_comp.rect

            self._resize(logical_pos, rect_data, self._orig_rpos, self._orig_rsizes, bounds)
        else:
            nodes = [n for n in self.nodes if not n.lockNode]
            rect_data = cast(List[RectData], self.compartments) + cast(List[RectData], nodes)
            if len(rect_data) == 0:
                return True
            self._move(logical_pos, rect_data, self._rel_positions)
        return True

    def _resize(self, pos: Vec2, rect_data: List[RectData], orig_pos: List[Vec2],
                orig_sizes: List[Vec2], bounds: Rect):
        """Helper that performs resize on the bounding box, given the logical mouse position."""
        # STEP 1, get new rect vertices
        # see class comment for resize handle format. For side-handles, get the vertex in the
        # counter-clockwise direction
        dragged_idx = self._resize_handle // 2
        # get the vertex opposite dragged idx as fixed_idx
        fixed_idx = int((dragged_idx + 2) % 4)
        orig_dragged_point = self._orig_rect.nth_vertex(dragged_idx)
        cur_dragged_point = self.outline_rect().nth_vertex(dragged_idx)
        fixed_point = self._orig_rect.nth_vertex(fixed_idx)

        target_point = pos + self._resize_handle_offset

        # if a side-handle, then only resize one axis
        if self._resize_handle % 2 == 1:
            if self._resize_handle % 4 == 1:
                # vertical resize; keep x the same
                target_point = target_point.swapped(0, orig_dragged_point.x)
            else:
                assert self._resize_handle % 4 == 3
                target_point = target_point.swapped(1, orig_dragged_point.y)

        # clamp target point
        target_point = clamp_point(target_point, bounds)

        # STEP 2, get and validate rect ratio

        # raw difference between (current/target) dragged vertex and fixed vertex. Raw as in this
        # is the visual difference shown on the bounding rect.
        orig_diff = orig_dragged_point - fixed_point
        target_diff = target_point - fixed_point

        signs = orig_diff.elem_mul(target_diff)

        # bounding_rect flipped?
        if signs.x < 0:
            target_point = target_point.swapped(0, cur_dragged_point.x)

        if signs.y < 0:
            target_point = target_point.swapped(1, cur_dragged_point.y)

        # take absolute value and subtract padding to get actual difference (i.e. sizing)
        pad_off = Vec2.repeat(self._padding)
        orig_bb_size = (orig_dragged_point -
                        fixed_point).elem_abs() - pad_off * 2
        target_size = (target_point - fixed_point).elem_abs() - pad_off * 2

        size_ratio = target_size.elem_div(orig_bb_size)

        # size too small?
        if size_ratio.x < self._min_resize_ratio.x:
            size_ratio = size_ratio.swapped(0, self._min_resize_ratio.x)

        if size_ratio.y < self._min_resize_ratio.y:
            size_ratio = size_ratio.swapped(1, self._min_resize_ratio.y)

        # re-calculate target_size in case size_ratio changed
        target_size = orig_bb_size.elem_mul(size_ratio)

        # STEP 3 calculate new bounding_rect position and size
        br_pos = Vec2(min(fixed_point.x, target_point.x), min(fixed_point.y, target_point.y))

        # STEP 4 calculate and apply new node positions and sizes
        # calculate and keep incremental ratio for the event arguments
        inc_ratio = orig_bb_size.elem_mul(size_ratio).elem_div(rect_data[0].size)
        offsets = list()
        index = 0
        for index, (rdata, opos, osize) in enumerate(zip(rect_data, orig_pos, orig_sizes)):
            #assert opos.x >= -1e-6 and opos.y >= -1e-6
            pos = (br_pos + opos.elem_mul(size_ratio) + pad_off)
            # HACK rect_data is compartments + nodes. So we only append to offsets we've reached
            # the nodes sectoin
            if index >= len(self.compartments):
                offsets.append(pos - rdata.position)
            rdata.position = pos
            rdata.size = osize.elem_mul(size_ratio)

        # STEP 5 apply new bounding_rect position and size
        self.bounding_rect.position = (br_pos + pad_off)
        self.bounding_rect.size = target_size

        # STEP 6 post main events
        if len(self.nodes) != 0:
            post_event(DidResizeNodesEvent(node_indices=self.node_indices, ratio=inc_ratio,
                                           dragged=True))
            post_event(DidMoveNodesEvent(node_indices=self.node_indices,
                                         offset=offsets[len(self.compartments):], dragged=True))
        if len(self.compartments) != 0:
            post_event(DidResizeCompartmentsEvent(
                compartment_indices=self.comp_indices, ratio=inc_ratio, dragged=True))
            post_event(DidMoveCompartmentsEvent(
                compartment_indices=self.comp_indices,
                offset=offsets[:len(self.compartments)], dragged=True))

        # STEP 7 adjust peripheral node positions so that they are inside the compartment
        adjusted_nodes = list()
        offsets = list()
        for node in self.peripheral_nodes:
            assert node.comp_idx != -1
            comp = self.canvas.comp_idx_map[node.comp_idx]
            pos = clamp_rect_pos(node.rect, comp.rect)
            if node.position != pos:
                adjusted_nodes.append(node)
                offsets.append(pos - node.position)
            node.position = pos
        if len(adjusted_nodes) != 0:
            post_event(DidMoveNodesEvent(adjusted_nodes, offsets, dragged=True))

    def _move(self, pos: Vec2, rect_data: List[RectData], rel_positions: List[Vec2]):
        """Helper that performs resize on the bounding box, given the logical mouse position."""
        assert len(rect_data) == len(rel_positions)
        assert len(rect_data) != 0
        # campute tentative new positions. May need to clamp it.
        self._did_move = True
        new_positions = [pos + rp for rp in rel_positions]
        min_x = min(p.x for p in new_positions)
        min_y = min(p.y for p in new_positions)
        max_x = max(p.x + r.size.x for p,
                    r in zip(new_positions, rect_data))
        max_y = max(p.y + r.size.y for p,
                    r in zip(new_positions, rect_data))
        offset = Vec2(0, 0)

        s_bounds = self._bounds
        lim_topleft = s_bounds.position
        lim_botright = s_bounds.position + s_bounds.size

        if min_x < lim_topleft.x:
            assert max_x <= lim_botright.x
            offset += Vec2(lim_topleft.x - min_x, 0)
        elif max_x > lim_botright.x:
            offset += Vec2(lim_botright.x - max_x, 0)

        if min_y < lim_topleft.y:
            assert max_y <= lim_botright.y
            offset += Vec2(0, lim_topleft.y - min_y)
        elif max_y > lim_botright.y:
            offset += Vec2(0, lim_botright.y - max_y)

        self.bounding_rect.position = (pos + offset + self._drag_rel)
        # The actual amount moved by the rects
        pos_offset = (new_positions[0] + offset) - rect_data[0].position
        for rdata, np in zip(rect_data, new_positions):
            rdata.position = (np + offset)

        # Note that don't need to test if out of bounds by peripheral nodes, since all
        # peripheral nodes must be inside selected compartments.
        for node in self.peripheral_nodes:
            node.position += pos_offset

        all_nodes = self.nodes + self.peripheral_nodes
        if len(all_nodes) != 0:
            post_event(DidMoveNodesEvent([n.index for n in all_nodes], pos_offset, dragged=True))
        if len(self.compartments) != 0:
            post_event(DidMoveCompartmentsEvent(compartment_indices=[c.index for c in self.compartments],
                                                offset=pos_offset, dragged=True))

    def move_offset(self, offset: Vec2):
        nodes = [n for n in self.nodes if not n.lockNode]
        rect_data = cast(List[RectData], self.compartments) + cast(List[RectData], nodes)
        pos = self.bounding_rect.position
        rel_node_pos = [n.position - pos for n in self.nodes if not n.lockNode]
        rel_comp_pos = [c.position - pos for c in self.compartments]
        rel_positions = rel_comp_pos + rel_node_pos

        if len(rect_data) == 0:
            return

        self._move(pos + offset, rect_data, rel_positions)
        self._commit_move()

    def _commit_move(self):
        with self.controller.group_action():
            for node in chain(self.nodes, self.peripheral_nodes):
                self.controller.move_node(self.net_index, node.index, node.position)

            for comp in self.compartments:
                self.controller.move_compartment(
                    self.net_index, comp.index, comp.position)

            if self.special_mode == SelectBox.SMode.NODES_IN_ONE:
                compi = self.canvas.InWhichCompartment(self.nodes)
                old_compi = self.nodes[0].comp_idx
                if compi != old_compi:
                    for node in self.nodes:
                        self.controller.set_compartment_of_node(
                            self.net_index, node.index, compi)
                    post_event(DidChangeCompartmentOfNodesEvent(
                        node_indices=[n.index for n in self.nodes],
                        old_compi=old_compi,
                        new_compi=compi
                    ))

            post_event(DidCommitDragEvent(self))

def primitive_peninfo(color: Color, width: float, is_alias: bool):
    info = wx.GraphicsPenInfo(color.to_wxcolour(), width)
    if is_alias:
        info = info.Style(wx.PENSTYLE_SHORT_DASH)
    return info


def primitive_brush(color: Color, is_alias: bool):
    # style = wx.BRUSHSTYLE_FIRST_HATCH if is_alias else wx.BRUSHSTYLE_SOLID
    style = wx.BRUSHSTYLE_SOLID
    brush = wx.Brush(color.to_wxcolour(), style=style)
    return brush


def draw_circle_to_gc(gc: wx.GraphicsContext, box: Rect, circle: CirclePrim, is_alias: bool):
    peninfo = primitive_peninfo(circle.border_color, circle.border_width, is_alias)
    pen = gc.CreatePen(peninfo)
    brush = gc.CreateBrush(primitive_brush(circle.fill_color, is_alias))
    gc.SetPen(pen)
    gc.SetBrush(brush)
    path = gc.CreatePath()
    path.AddEllipse(box.position.x, box.position.y, box.size.x, box.size.y)
    gc.DrawPath(path)


def draw_rect_to_gc(gc: wx.GraphicsContext, box: Rect, rect: RectanglePrim, is_alias: bool):
    # restrict the border width of the node to an even integer
    # this is necessary for aligning the node rectangle to the selection rectangle
    # why? see Rect::aligned().
    box = box.aligned()
    if rect.border_width == 0:
        border_width = 0
    else:
        border_width = max(even_round(rect.border_width), 2)
    pen = gc.CreatePen(primitive_peninfo(rect.border_color, border_width, is_alias))
    brush = gc.CreateBrush(primitive_brush(rect.fill_color, is_alias))
    gc.SetPen(pen)
    gc.SetBrush(brush)
    gc.DrawRoundedRectangle(box.position.x, box.position.y, box.size.x, box.size.y, rect.corner_radius)


def draw_polygon_to_gc(gc: wx.GraphicsContext, box: Rect, poly: PolygonPrim, is_alias: bool):
    pen = gc.CreatePen(primitive_peninfo(poly.border_color, poly.border_width, is_alias))
    brush = gc.CreateBrush(primitive_brush(poly.fill_color, is_alias))
    gc.SetPen(pen)
    gc.SetBrush(brush)
    # offset so polygon is centered at (0.5, 0.5)
    points = [p + Vec2.repeat(0.5) for p in poly.points]
    transformed_pts = [box.position + p.elem_mul(box.size) for p in points]
    gc.DrawLines([wx.Point2D(*p) for p in transformed_pts])


# HACK bypass Python typing warnings with Type variables
draw_fn_map: Dict[Any, Any] = {
    CirclePrim: draw_circle_to_gc,
    RectanglePrim: draw_rect_to_gc,
    HexagonPrim: draw_polygon_to_gc,
    LinePrim: draw_polygon_to_gc,
    TrianglePrim: draw_polygon_to_gc
}


# def apply_transform_to_gc(gc: wx.GraphicsContext, transform: Transform):
#     gc.Translate(*transform.translation)
#     gc.Rotate(transform.rotation)
#     gc.Scale(*transform.scale)


def draw_composite_shape(gc: wx.GraphicsContext, bounding_rect: Rect, node: Node):
    shape = node.composite_shape
    gc.PushState()
    # gc.Translate(*(bounding_rect.position.elem_mul(bounding_rect.size)))
    # gc.Scale(*bounding_rect.size)
    for primitive, transform in shape.items:
        gc.PushState()
        # apply_transform_to_gc(gc, transform)
        box = Rect(bounding_rect.position + bounding_rect.size.elem_mul(transform.translation),
                   bounding_rect.size.elem_mul(transform.scale))
        draw_fn_map[primitive.__class__](gc, box, primitive, node.original_index != -1)
        gc.PopState()
    gc.PopState()

    gc.PushState()
    # apply_transform_to_gc(gc, node.composite_shape.text_item[1])
    draw_text_to_gc(gc, bounding_rect, node.id, node.composite_shape.text_item)
    gc.PopState()


def _truncate_text(gc: wx.GraphicsContext, max_width: float, text: str):
    '''Truncate the given text (add ellipsis to the end) so that its width fits inside `max_width`.

    Assume that we are using the currently selected font in gc.
    '''

    tw, th, _, _ = gc.GetFullTextExtent(text)
    # very rough chopping
    if tw > max_width:
        text_len = int((max_width / tw) * len(text))
        text = text[:text_len]
        if len(text) > 4:
            text = text[:-3] + '....'
        else:
            text = '....'

    return text


def draw_text_to_gc(gc: wx.GraphicsContext, bounding_rect: Rect, full_text_string, text_item: Tuple[TextPrim, Transform]):
    primitive, transform = text_item

    # Maybe cache this?
    fg_color = primitive.font_color.to_wxcolour()
    font = wx.Font(wx.FontInfo(primitive.font_size)
                   .Family(primitive.font_family)
                   .Style(primitive.font_style)
                   .Weight(primitive.font_weight))
    gfont = gc.CreateFont(font, fg_color)
    gc.SetFont(gfont)
    brush = gc.CreateBrush(wx.Brush(primitive.bg_color.to_wxcolour()))

    width, height = bounding_rect.size
    text_string = _truncate_text(gc, bounding_rect.size.x, full_text_string)
    tw, th, _, _ = gc.GetFullTextExtent(text_string)
    tw_full, th_full, _, _ = gc.GetFullTextExtent(full_text_string)

    # remaining x and y
    rx = width - tw
    ry = height - th
    # upper left corner of node
    text_pos = bounding_rect.position + transform.translation.elem_mul(bounding_rect.size)
    # do not truncate text if outside node
    if not primitive.position == TextPosition.IN_NODE:
        text_string = full_text_string
        text_pos = text_pos - Vec2(tw_full + th/2, 0)
        rx = th + tw_full + width

    if primitive.alignment == TextAlignment.LEFT:
        draw_pos = text_pos + Vec2(0, ry / 2)
    elif primitive.alignment == TextAlignment.CENTER:
        draw_pos = text_pos + Vec2(rx, ry) / 2
    elif primitive.alignment == TextAlignment.RIGHT:
        draw_pos = text_pos + Vec2(rx, ry / 2)
    else:
        assert False, "This should not happen"

    if primitive.position == TextPosition.ABOVE:
        draw_pos = draw_pos + Vec2(0, -height/2 - 2*th)
    elif primitive.position == TextPosition.BELOW:
        draw_pos = draw_pos + Vec2(0, height/2 + 2*th)

    gc.DrawText(text_string, draw_pos.x, draw_pos.y, brush)




'''
Given a random network, this plugin  will rearrange the network neatly on the screen.
Version 1.0.2: Author: Carmen Perena Cortes, Herbert M Sauro, Jin Xu, 2022
Based on THOMAS M. J. FRUCHTERMAN AND EDWARD M. REINGOLD's Graph Drawing by Force-directed Placement
SOFTWARE - PRACTICE AND EXPERIENCE, VOL. 21(1 1), 1129-1164 (NOVEMBER 1991)
Using python's networkx
'''
# pylint: disable=maybe-no-member
import wx
from rkviewer.plugin.classes import PluginMetadata, WindowedPlugin, PluginCategory
from rkviewer.plugin import api
from rkviewer.plugin.api import Node, Vec2, Reaction, get_reaction_by_index
import math
from dataclasses import field
from typing import List
from random import uniform
from numpy.random import rand
import numpy as np
import random
from dataclasses import dataclass
import networkx as nx
import rkviewer.canvas.utils as cu

class LayoutNetworkX(WindowedPlugin):
    metadata = PluginMetadata(
        name='AutoLayout',
        author='Carmen Perena Cortes, Herbert M Sauro and Jin Xu',
        version='1.0.2',
        short_desc='Auto Layout using networkX.',
        long_desc='Rearrange a random network into a neat auto layout',
        category=PluginCategory.UTILITIES,
    )

    def create_window(self, dialog):
        '''
        Create a window with several inputs and buttons.
        Args:
            self
            dialog
        '''
        # TODO: k, gravity, useMagnetism, useBoundary, useGrid
        self.window = wx.Panel(dialog, pos=(5,100), size=(350, 220))
        self.sizer = wx.FlexGridSizer(cols=2, vgap=10, hgap=0)
        self.sizer.AddGrowableCol(0, int(0.6))
        self.sizer.AddGrowableCol(1, int(0.4))

        self.sizer.Add((0, 10))
        self.sizer.Add((0, 10))

        self.MaxIterText = wx.TextCtrl(self.window, -1, "100", size=(100, -1))
        self.MaxIterText.SetInsertionPoint(0)
        self.MaxIterText.Bind(wx.EVT_TEXT, self.OnText_MaxIter)
        self.MaxIterValue = int(self.MaxIterText.GetValue())
        self.AddField('Max Number of Iterations', self.MaxIterText)

        self.kText = wx.TextCtrl(self.window, -1, "70", size=(100, -1))
        self.kText.SetInsertionPoint(0)
        self.kText.Bind(wx.EVT_TEXT, self.OnText_k)
        self.kValue = float(self.kText.GetValue())
        self.AddField('k (float > 0)', self.kText)

        self.scaleText = wx.TextCtrl(self.window, -1, "550", size=(100, -1))
        self.scaleText.SetInsertionPoint(0)
        self.scaleText.Bind(wx.EVT_TEXT, self.OnText_scale)
        self.scaleValue = float(self.scaleText.GetValue())
        self.AddField('Scale of Layout', self.scaleText)

        self.useCentroid = False
        self.centroidCheckBox = wx.CheckBox(self.window)
        self.centroidCheckBox.Bind(wx.EVT_CHECKBOX, self.OnCheckUseCentroid)
        self.AddField('Also Arrange Centroids', self.centroidCheckBox)

        # add spacer left of button
        self.sizer.Add((0, 0))
        apply_btn = wx.Button(self.window, -1, 'Run', (220, 130))
        apply_btn.Bind(wx.EVT_BUTTON, self.Apply)
        self.sizer.Add(apply_btn)
        self.window.SetPosition (wx.Point(10,10))
        self.window.SetSizer(self.sizer)
        return self.window

    def AddField(self, text, field):
        self.sizer.Add(wx.StaticText(self.window, label=text), wx.SizerFlags().Border(wx.LEFT, 20))
        self.sizer.Add(field)

    def OnText_MaxIter(self, evt):
        try:
          update = evt.GetString()
          if update != '':
              self.MaxIterValue = int(self.MaxIterText.GetValue())
        except ValueError:
           wx.MessageBox('Value must be a number', 'Error', wx.OK | wx.ICON_INFORMATION)

    def OnText_k(self, evt):
        try:
           update = evt.GetString()
           if update != '':
              self.kValue = float(self.kText.GetValue())
        except ValueError:
           wx.MessageBox('Value must be a number', 'Error', wx.OK | wx.ICON_INFORMATION)

    def OnText_scale(self, evt):
        try:
           update = evt.GetString()
           if update != '':
              self.scaleValue = float(self.scaleText.GetValue())
        except ValueError:
           wx.MessageBox('Value must be a number', 'Error', wx.OK | wx.ICON_INFORMATION)

    def OnCheckUseCentroid(self, evt):
        cb = evt.GetEventObject()
        self.useCentroid = cb.GetValue()

    def Apply(self, evt):
        if api.node_count(0) == 0:
            return
        G = nx.Graph()
        nodesInd = np.array(list(api.get_node_indices(0)))
        reactionsInd =  np.array(list(api.get_reaction_indices(0)))
        originalPos = {}
        def generateGraph():
            # add nodes and centroids as "nodes" for networkX
            nodes = np.array(list(api.get_node_indices(0)))
            #reactionsInd =  np.array(list(api.get_reaction_indices(0)))
            cStr = np.empty_like(reactionsInd, dtype=str)
            cStr[:,] = "c"
            centroidId = np.char.add(cStr, reactionsInd.astype(str))
            G.add_nodes_from(centroidId)
            nStr = np.empty_like(nodes, dtype=str)
            nStr[:,] = "n"
            nodesId = np.array(list(np.char.add(nStr, nodesInd.astype(str))))
            G.add_nodes_from(nodesId)
            '''
            for n in nodes:
                centroidsTo = np.array(list(api.get_reactions_as_product(0, n))) # Gets the reactions in which it is a product -> TO
                cStr = np.empty_like(centroidsTo, dtype=str)
                cStr[:,] = "c"
                centroidsToIn = np.char.add(cStr, centroidsTo.astype(str)) # centroids from which the node is product
                centroidsFrom = np.array(list(api.get_reactions_as_reactant(0, n))) # reactions in which it is a reactanr -> FROM
                cStr = np.empty_like(centroidsFrom, dtype=str)
                cStr[:,] = "c"
                centroidsFromIn = np.char.add(cStr, centroidsFrom.astype(str)) # centroids to which the node is reactant
                nS = np.empty_like(centroidsToIn, dtype = str)
                nS[:,] = "n"
                numS = np.empty_like(centroidsToIn, dtype = int)
                numS[:,] = n
                nodeIndArrayTo = np.char.add(nS, numS.astype(str))
                nS = np.empty_like(centroidsFromIn, dtype = str)
                nS[:,] = "n"
                numS = np.empty_like(centroidsFromIn, dtype = int)
                numS[:,] = n
                nodeIndArrayFrom = np.char.add(nS, numS.astype(str))
                edgesTo = np.array(list(zip(centroidsToIn, nodeIndArrayTo)))
                edgesFrom = np.array(list(zip(nodeIndArrayFrom, centroidsFromIn)))
                G.add_edges_from(edgesTo)
                G.add_edges_from(edgesFrom)
            '''

            # Add edges from reactant to centroid and centroid to product (undirected)
            edges = list()
            for reaction in api.get_reactions(0):
                for s in reaction.sources:
                    edges.append(('n' + str(s), 'c' + str(reaction.index)))
                for t in reaction.targets:
                    edges.append(('c' + str(reaction.index), 'n' + str(t)))

            G.add_edges_from(edges)
            cn = 0
            for rea in api.get_reactions(0):
                cent = api.compute_centroid(0, rea.sources, rea.targets)
                #originalPos[centroidId[cn]] = list([cent.x, cent.y])
                originalPos['c' + str(rea)] = list([random.randint(0,600), random.randint(0,600)])
                cn = cn + 1

            for nod in api.get_nodes(0):
                #originalPos[nodesId[cn]] = list([nod.position.x, nod.position.y])
                # random.randint(0,500), nod.position.y+random.randint (0,500)])
                originalPos['n' + str(nod)] = list([random.randint(0,600), random.randint (0,600)])
                cn = cn + 1
        generateGraph()
        #print(nx.to_dict_of_lists(G))
        #nodeIds = list (api.get_node_indices(0))
        with api.group_action():
            for t in range(1):
                pos = (nx.fruchterman_reingold_layout(G, k = self.kValue, iterations = self.MaxIterValue, scale = self.scaleValue, pos = originalPos, weight=1))
                positions = np.array(list(pos.values()))
                minX = 0
                minY = 0
                for p in positions:
                    if p[0] < minX:
                        minX = p[0]
                    if p[1] < minY:
                        minY = p[1]
                positions = positions - np.array([minX, minY])
                centroids = positions[0: len(reactionsInd)]
                nodes = positions[len(reactionsInd): len(positions)]
                count = 0
                for n in nodes:
                    newX = float(n[0])
                    newY = float(n[1])
                    api.move_node(0, nodesInd[count], position = Vec2(newX, newY), allowNegativeCoordinates=True)
                    count = count + 1

                if self.useCentroid:
                    count = 0
                    for c in centroids:
                        newX = float(c[0])
                        newY = float(c[1])
                        # r = api.get_reaction_by_index(0, reactionsInd[count])
                        # handles = api.default_handle_positions(0, r.index)
                        # api.update_reaction(0, r.index, center_pos=Vec2(newX, newY), handle_positions=handles)
                        index = reactionsInd[count]
                        r = api.get_reaction_by_index(0, index)
                        rcts = r.sources
                        prds = r.targets
                        centroid = Vec2(newX, newY)
                        #update centroid before calculate the default handles
                        # api.update_reaction(0, index, center_pos=Vec2(newX, newY))
                        # handles = api.default_handle_positions(0, r.index)
                        # api.update_reaction(0, index, handle_positions=handles)

                        #set the handles as centroid to make all the bezier curves look like straight lines
                        handles = [centroid]
                        for x in range(len(rcts)):
                            handles.append(centroid)
                        for y in range(len(prds)):
                            handles.append(centroid)
                        api.update_reaction(0, index, center_pos=Vec2(newX, newY), handle_positions=handles)
                        
                        count = count + 1
                else:
                    for index in api.get_reaction_indices(0):
                        #api.update_reaction(0, index, center_pos=None)
                        #handles = api.default_handle_positions(0, index)
                        #api.update_reaction(0, index, handle_positions=handles)
                        #the following centroid computed is the same as default centroid, or None
                        rcts = get_reaction_by_index(0, index).sources
                        prds = get_reaction_by_index(0, index).targets
                        centroid = api.compute_centroid(0, rcts, prds)
                        #set the handles as centroid to make all the bezier curves look like straight lines
                        handles = [centroid]
                        for x in range(len(rcts)):
                            handles.append(centroid)
                        for y in range(len(prds)):
                            handles.append(centroid)
                        api.update_reaction(0, index, center_pos = centroid, handle_positions = handles)

                '''

                for r in api.get_reactions(0):
                    currCentroid = centroids[r.index]
                    newX = float(currCentroid[0])
                    newY = float(currCentroid[1])
                    api.update_reaction(0, r.index, center_pos=(Vec2(newX, newY)))
                    #api.update_reaction(0, r.index, center_pos=None)
                    handles = api.default_handle_positions(0, r.index)
                    api.set_reaction_center_handle(0, r.index, handles[0])
                    count = 1
                    for s in r.sources:
                        api.set_reaction_node_handle(0, r.index, s, True, handles[count])
                        count += 1
                    for t in r.targets:
                        api.set_reaction_node_handle(0, r.index, t, False, handles[count])
                        count += 1

                '''

            ws = api.window_size()
            offset = Vec2(ws.x/4, ws.y/4)
            api.translate_network(0, offset, check_bounds = True)


"""Configuration parameters.
"""
# pylint: disable=maybe-no-member
# from __future__ import annotations

import copy
import os
from dataclasses import dataclass
from pathlib import Path
from typing import Any, List, Tuple

import commentjson
import wx
from commentjson.commentjson import JSONLibraryException
from marshmallow import Schema, ValidationError, fields
from marshmallow import missing as missing_
from marshmallow import validate

from rkviewer.canvas.geometry import Rect, Vec2
from rkviewer.utils import get_local_path

# TODO rename get_local_path
#config_dir = "" #get_local_path('.rkviewer')  # This will be set by either appsettings (see below) or CreateConfigDir in view.poy
#settings_theme_path = os.path.join(config_dir, 'settings.json')
#default_settings_path = os.path.join(config_dir, '.default-settings.json')


def GetConfigDir ():
    sp = wx.StandardPaths.Get()
    return sp.GetUserConfigDir()


def GetThemeSettingsPath ():
    sp = wx.StandardPaths.Get()
    config_dir = sp.GetUserConfigDir()
    return os.path.join(config_dir, 'rkViewer', 'settings.json')


@dataclass(frozen=True, eq=True)
class Color:
    r: int
    g: int
    b: int
    a: int = 255

    def swapped(self, r: int = None, g: int = None, b: int = None, a: int = None) -> 'Color':
        if r is None:
            r = self.r
        if g is None:
            g = self.g
        if b is None:
            b = self.b
        if a is None:
            a = self.a
        return Color(r, g, b, a)

    def to_wxcolour(self):
        return wx.Colour(self.r, self.g, self.b, self.a)

    @classmethod
    def from_rgb(cls, val: int) -> 'Color':
        """Create color from RGB hex value: #00BBGGRR"""
        return Color((val >> 0) & 255, (val >> 8) & 255, (val >> 16) & 255)



@dataclass
class Font:
    pointSize: int
    color: Color
    # family: str  # TODO change to enum
    # style: str
    # weight: str
    # name: str

    # def __init__(self):
    #     self.pointSize = 20
    #     self.color = Color(0, 0, 0, 255)
    #     #self.family = "default"
    #     #self.style = "normal"
    #     #self.weight = "default"
    #     #self.name = ""


# Application based settings stored in an ini file
class AppSettings:

   # Default positions

   def __init__(self):
       self.position = wx.Point (0,0)
       self.size = wx.Size (1366,737)
       self.displaySize = wx.DisplaySize()
       # Find the center poistion of the frame
       self.position.x = self.displaySize[0] // 2 - self.size.x // 2
       self.position.y = self.displaySize[1] // 2 - self.size.y // 2

   def load_appSettings(self):
       sp = wx.StandardPaths.Get()
       configDir = sp.GetUserConfigDir()
       if not os.path.exists(os.path.join(configDir, 'rkViewer')):
            os.mkdir(os.path.join(configDir, 'rkViewer'))
       fileName = os.path.join(configDir, 'rkViewer', 'appSettings.ini')
       if not os.path.exists (fileName):
          # leave with defaults intact.
          return

       config = wx.FileConfig(localFilename=fileName)
       config.SetPath ('PositionAndSize')
       self.position.x = config.ReadInt ('frame_position_x', self.position.x)
       self.position.y = config.ReadInt ('frame_position_y', self.position.y)
       self.size.x = config.ReadInt ('frame_size_w', self.size.x)
       self.size.y = config.ReadInt ('frame_size_h', self.size.y)

   def save_appSettings(self):
       sp = wx.StandardPaths.Get()
       config_dir = sp.GetUserConfigDir()
       if not os.path.exists(os.path.join(config_dir, 'rkViewer')):
            os.mkdir(os.path.join(config_dir, 'rkViewer'))
       fileName = os.path.join(config_dir, 'rkViewer', 'appSettings.ini')
       if os.path.exists(fileName):
           os.remove (fileName)

       config = wx.FileConfig(localFilename=fileName)
       config.SetPath ('PositionAndSize')
       config.WriteInt ('frame_position_x', self.position.x)
       config.WriteInt ('frame_position_y', self.position.y)
       config.WriteInt ('frame_size_w', self.size.x)
       config.WriteInt ('frame_size_h', self.size.y)


# TODO merge these schema with those in iodine?
class ColorField(fields.Field):
    """Field that represents an RGBA color.

    To represent the color red, you would write:
    >>> { "some_color": [255, 0, 0] }

    You may also specify its opacity. To make the color red half transparent:
    >>> { "some": [255, 0, 0, 127] }

    In short, you may specify four integer arguments RGBA in an array, which the alpha value
    being optional and defaulting to 255, or fully opaque. Each value must be in range [0, 255].
    """
    list_field = fields.List(fields.Int(), validate=validate.Length(min=3, max=4))
    range_validate = validate.Range(min=0, max=255, error='RGBA values must be between 0 and 255.')

    def __init__(self, **kw):
        super().__init__(**kw)

    def _serialize(self, value: Color, attr, obj, **kwargs):
        ret = [value.r, value.g, value.b]
        if value.a != 255:
            ret += [value.a]
        return ret

    def _deserialize(self, value, attr, data, **kwargs) -> Color:
        self.list_field.validate(value)
        for val in value:
            ColorField.range_validate(val)
        return Color(*value)


class Pixel(fields.Int):
    """Field that represents some length in pixels.

    The only current restriction is that this must be a nonnegative integer, or
    >>> { "some_width": 23 }
    """
    def __init__(self, **kwargs):
        super().__init__(validate=validate.Range(min=0), **kwargs)


class Dim(fields.Float):
    """Field that represents some real dimension (length)."""

    def __init__(self, **kwargs):
        # TODO should we allow 0? Also decide for pixel
        super().__init__(validate=validate.Range(min=0), **kwargs)


class Dim2(fields.List):
    def __init__(self, *args, **kw):
        super().__init__(Dim(), *args, validate=validate.Length(equal=2), **kw)

    def _serialize(self, value: Vec2, attr, obj, **kwargs):
        if value is None:
            return None
        return (value.x, value.y)

    def _deserialize(self, value, attr, data, **kwargs):
        self._validate(value)
        return Vec2(value)


class FontField(fields.Field):
    def __init__(self, *args, **kw):
        pass

    def _serialize(self, value: Font, attr, obj, **kwargs):
        return vars(value)

    def _validate(self, value):
        print('hi')  # TODO
        pass

    def _deserialize(self, value, attr, data, **kwargs):
        return Font(**value)


class ThemeSchema(Schema):
    """Schema for the overall theme, i.e. appearance, of the application.

    Attributes:
        overall_bg: Overall background color of the application.
        canvas_bg: Background color of the canvas.
        toolbar_bg: Background color of the toolbar.
        btn_bg : Background color of a button
        btn_fg: Foreground color of a button, ie the color used by the button text
        btn_hover_bg: BG color of a button when the mouse is hovering over it. By default
                      this should be equal to the normal BG since MSW already provides a faint
                      blue hovering effect. But in the case that btn_border is False, the native
                      hovering effects vanish, so this is needed as a replacement visual indicator.
        btn_hover_fg: FG color of a button when the mouse is hovering over it. See btn_hover_bg.
        btn_border: True for border, false for no border
        canvas_width: Displayed (visible) width of the canvas.
        real_canvas_width: Actual width of the canvas, some of which may be hidden
        canvas_height: Displayed (visible) height of the canvas.
        real_canvas_height: Actual height of the canvas, some of which may be hidden
        vgap: Vertical gap between toolbars and canvas.
        hgap: Horizontal gap between toolbars and canvas.
        canvas_outside_bg: Background color of the part outside of the bounds of canvas.
        mode_panel_width: Width of the mode selection panel.
        node_fill: Default node fill color.
        node_border: Default node border color.
        node_width: Default node width.
        node_height: Default node height.
        node_border_width: Default node border width.
        node_font_size: Default font size of the node.
        node_font_color: Default font color of the node.
        select_outline_width: Width of the selection outlines (i.e. outline around each selected item).
        select_outline_padding: Padding of the selection outlines.
        handle_color: ColorField of the Reaction Bezier curves.
        highlighted_handle_color: ColorField of the Reaction Bezier curves when the cursor hovers over it
        select_box_padding: Padding of the selection rectangle (i.e. the large rectangle that
                            encompasses all the selected items).
        select_handle_length: Side length of the squares one uses to resize nodes/compartments.
        zoom_slider_bg: Background color of the zoom slider.
        drag_fill: The fill color of the drag selection rectangle.
        drag_border: The border color of the drag selection rectangle.
        drag_border_width: The border width of the drag selection rectangle.
        react_node_padding: The outline padding for nodes prepped as reactants or products.
        react_node_border_width: The border width for nodes prepped as reactants or products.
        reactant_border: The outline color for nodes prepped as reactants.
        product_border: The outline color for nodes prepped as products.
        reaction_fill: The default fill color of reaction curves.
        reaction_line_thickness: The default thickness of reaction curves.
        selected_reaction_fill: The fill color of selected reaction curves.
        comp_fill: The default fill color of compartments.
        comp_border: The default border color of compartments.
        comp_border_width: The default border width of compartments.
        reaction_radius: The radius of the reaction centroid circles.
        active_tab_fg: The foreground (font) color of the active tab in a FlatNotebook. There is
                       no option to set the BG color since the BG color is always a brighter version
                       of the toolbar BG.
        text_field_bg: The background color of text fields.
        text_field_fg: The foreground (font) color of text fields
        text_field_border: Whether the border of text fields should be drawn.

    TODO more documentation under attributes and link to this document in Help or settings.json
    """
    # overall background of the application
    overall_bg = ColorField(missing=Color(240, 240, 240))
    canvas_bg = ColorField(missing=Color(255, 255, 255))
    # Background color of the toolbars (i.e. panels around the canvas)
    toolbar_bg = ColorField(missing=Color(230, 230, 230))
    # Text color of the toolbars (i.e. panels around the canvas)
    toolbar_fg = ColorField(missing=Color(0, 0, 0))
    canvas_width = Pixel(missing=1000)
    real_canvas_width = Pixel(missing=10000)
    canvas_height = Pixel(missing=620)
    real_canvas_height = Pixel(missing=6200)
    btn_bg = ColorField(missing=Color(240,240,240))
    btn_fg = ColorField(missing=Color(0, 0, 0))
    btn_hover_bg = ColorField(missing=Color(240,240,240))
    btn_hover_fg = ColorField(missing=Color(0, 0, 0))
    btn_border = fields.Boolean(missing=True)
    # vertical gap between toolbars and canvas
    vgap = Pixel(missing=2)
    # horizontal gap between toolbars and canvas
    hgap = Pixel(missing=2)
    canvas_outside_bg = ColorField(missing=Color(160, 160, 160))
    mode_panel_width = Pixel(missing=100)
    toolbar_height = Pixel(missing=75)
    edit_panel_width = Pixel(missing=260)
    node_fill = ColorField(missing=Color(255, 204, 153, 200))
    node_border = ColorField(missing=Color(255, 108, 9))
    node_width = Dim(missing=50)
    node_height = Dim(missing=30)
    # node_corner_radius = Dim(missing=0.15)
    node_border_width = Dim(missing=2)
    node_font_size = Pixel(missing=10)
    node_font_color = ColorField(missing=Color(255, 0, 0, 100))
    # Width of the outline around each selected node
    select_outline_width = Pixel(missing=2)
    # Padding of the outline around each selected node
    select_outline_padding = Pixel(missing=3)
    # ColorField of control handles, e.g. resize handles, Bezier handles
    handle_color = ColorField(missing=Color(0, 140, 255))
    # ColorField of control handles when they are highlighted, if applicable
    highlighted_handle_color = ColorField(missing=Color(128, 198, 255))
    # Padding of the select box, relative to the mininum possible bbox
    select_box_padding = Pixel(missing=5)
    # Length of the squares one uses to drag resize nodes
    select_handle_length = Pixel(missing=8)
    zoom_slider_bg = ColorField(missing=Color(180, 180, 180))
    drag_fill = ColorField(missing=Color(0, 140, 255, 20))
    drag_border = ColorField(missing=Color(0, 140, 255))
    drag_border_width = Pixel(missing=1)
    react_node_padding = Pixel(missing=5)
    react_node_border_width = Pixel(missing=3)
    reactant_border = ColorField(missing=Color(255, 100, 100))
    product_border = ColorField(missing=Color(0, 214, 125))
    reaction_fill = ColorField(missing=Color(91, 176, 253))
    reaction_line_thickness = Dim(missing=2)
    selected_reaction_fill = ColorField(missing=Color(0, 140, 255))
    comp_fill = ColorField(missing=Color(158, 169, 255, 200))
    comp_border = ColorField(missing=Color(0, 29, 255))
    comp_border_width = Dim(missing=2)
    comp_corner_radius = Dim(missing=6)
    reaction_radius = Dim(missing=6)
    modifier_line_color = ColorField(missing=Color(202, 148, 255))
    modifier_line_width = Dim(missing=2)
    active_tab_fg = ColorField(missing=Color(0, 0, 0))
    text_field_bg = ColorField(missing=Color(255, 255, 255))
    text_field_fg = ColorField(missing=Color(0, 0, 0))
    text_field_border = fields.Boolean(missing=True)
    # using ribbon style for notebooks, so no need for this
    # active_tab_bg = ColorField(missing=Color(100, 100, 100))
    # panel_font = FontField(missing=Font(pointSize=))


class RootSchema(Schema):
    """The overall root schema.

    Attributes:
        theme: The theme settings (i.e. colors and dimensions) of the application.
    """
    theme = fields.Nested(ThemeSchema, missing=ThemeSchema().load({}))


# TODO put this in the schema somewhere
DEFAULT_ARROW_TIP = [Vec2(1, 15), Vec2(4, 8), Vec2(1, 1), Vec2(21, 8)]


# These settings are not meant to be modifiable by the user
BUILTIN_SETTINGS = {
    "init_zoom": 1,
    "status_fields": [
        ("mode", 100),
        ("cursor", 100),
        ("zoom", 100),
        ("fps", 100),
    ],  # first element: status field identifier; second element: field width
    "decimal_precision": 2,
    "min_node_width": 20,
    "min_node_height": 15,
    "min_comp_width": 350,
    "min_comp_height": 200,
}


INIT_SETTING_TEXT = '''// settings.json auto-generated by RKViewer.
// Edit this file to override default application settings.
// After you have finished editing, click File > Reload Settings to load the new settings.
// Note that some settings may not take effect until the application is restarted.
{
    // Theme and appearance settings
    "theme": {

    }
}

'''


DEFAULT_SETTING_FMT = '''// default-settings.json auto-generated by RKViewer.
// Note that you should *not* edit this file, since this is only for reference on the system
// default settings, and this will overwritten each time "Default Settings..." is clicked.
// If you wish to override these settings, go to // File > Edit Settings... instead.
//
// The settings generated here may change depending on the loaded plugins, which have their own
// default settings.
{}

'''


root_schema = RootSchema()
def validate_schema(schema):
    for field in schema.fields.values():
        # make sure fields have a default replacement for missing fields
        assert field.missing != missing_


_settings = BUILTIN_SETTINGS
_theme = None
_settings_err = None



@dataclass
class RuntimeVariables:
    '''Variables pertaining to the application runtime'''
    enable_plugins: bool = True


_RUNTIME_VARS = RuntimeVariables()


def runtime_vars() -> RuntimeVariables:
    return _RUNTIME_VARS


def reset_runtime_vars():
    global _RUNTIME_VARS
    _RUNTIME_VARS = RuntimeVariables()


def load_theme_settings():
    """Reload all settings from the default settings path.

    The exceptions are not immediately thrown since the first time the settings are loaded, the
    app has not been initialized. So we wait until it is, and then display an error dialog if there
    is a previously recorded error.
    """
    global _theme, _settings, _settings_err
    cur_settings = dict()
    if os.path.isfile(GetThemeSettingsPath ()):
        with open(GetThemeSettingsPath (), 'r') as fp:
            try:
                cur_settings = commentjson.load(fp)
            except JSONLibraryException as e:
                _settings_err = e
                if _theme is not None:
                    # theme has already been loaded, so don't modify it. Otherwise load all
                    # defaults.
                    return
    try:
        temp = root_schema.load(cur_settings)
        config = temp
    except ValidationError as e:
        _settings_err = e
        if _theme is not None:
            return
        config = root_schema.load({})
    _theme = copy.copy(config['theme'])
    _settings = copy.copy(BUILTIN_SETTINGS)


def pop_settings_err():
    """If there was error when loading the settings JSON, this would be not None. Used by canvas to
    display a warning message after the window is created.

    The error is cleared when this is called.
    """
    global _settings_err
    ret = _settings_err
    _settings_err = None
    return ret


def get_default_raw_settings():
    return root_schema.dump(root_schema.load({}))


def get_setting(setting_attr) -> Any:
    return _settings[setting_attr]


def get_theme(theme_attr, convert_color=True) -> Any:
    '''convert_color: if True, convert Color instances to wx.Colour instances automatically
    '''
    global _theme
    tmp = _theme[theme_attr]
    # automatically convert Color to wx.Colour
    if convert_color and isinstance(tmp, Color):
        tmp = wx.Colour(tmp.r, tmp.g, tmp.b, tmp.a)
    return tmp


def add_plugin_schema(name: str, schema: Schema):
    # TODO
    pass




# pylint: disable=maybe-no-member
import wx
from rkviewer.config import DEFAULT_ARROW_TIP
from rkviewer.canvas.geometry import Rect, Vec2
import copy
from dataclasses import dataclass
from enum import Enum, unique
from typing import Any, Callable, List, Tuple

@unique
class InputMode(Enum):
    """Enum for the current input mode of the canvas."""
    SELECT = 'Select'
    ADD_NODES = 'Add Nodes'
    ADD_COMPARTMENTS = 'Add Compartments'
    ZOOM = 'Zoom'

    def __str__(self):
        return str(self.value)


class ArrowTip:
    points: List[Vec2]

    def __init__(self, points: List[Vec2]):
        if len(points) != 4:
            raise ValueError('Arrow tip must consist of 4 points!')
        self.points = points

    def clone(self):
        return ArrowTip(copy.copy(self.points))


@dataclass
class CanvasState:
    """The current global state of the canvas.

    Attributes:
        scale: The zoom scale of the canvas.
        multi_select: Whether the user is pressing the keys that signify multiple selection of
                      items.
    """
    scale: float = 1
    bounds: Rect = Rect(Vec2(), Vec2())
    input_mode_changed: Callable[[InputMode], None] = lambda _: None
    _input_mode: InputMode = InputMode.SELECT
    arrow_tip: ArrowTip = ArrowTip(copy.copy(DEFAULT_ARROW_TIP))

    @property
    def input_mode(self):
        return self._input_mode

    @input_mode.setter
    def input_mode(self, mode: InputMode):
        self._input_mode = mode
        self.input_mode_changed(mode)
    
    @property
    def multi_select(self):
        return wx.GetKeyState(wx.WXK_CONTROL) or wx.GetKeyState(wx.WXK_SHIFT)

cstate = CanvasState()



"""Utility functions for the canvas.

This includes drawing helpers and 2D geometry functions.
"""
# pylint: disable=maybe-no-member
import wx
import abc
import math
from typing import Collection, Generic, List, Optional, Set, TypeVar, Callable
from .geometry import Rect, Vec2, rotate_unit
from .data import Node, Reaction


def get_nodes_by_idx(nodes: List[Node], indices: Collection[int]):
    """Simple helper that maps the given list of indices to their corresponding nodes."""
    ret = [n for n in nodes if n.index in indices]
    assert len(ret) == len(indices)
    return ret

def get_rxns_by_idx(rxns: List[Reaction], indices: Collection[int]):
    """Simple helper that maps the given list of indices to their corresponding rxns."""
    ret = [n for n in rxns if n.index in indices]
    assert len(ret) == len(indices)
    return ret

def get_nodes_by_ident(nodes: List[Node], ids: Collection[str]):
    """Simple helper that maps the given list of IDs to their corresponding nodes."""
    ret = [n for n in nodes if n.id in ids]
    assert len(ret) == len(ids)
    return ret


def draw_rect(gc: wx.GraphicsContext, rect: Rect, *, fill: Optional[wx.Colour] = None,
              border: Optional[wx.Colour] = None, border_width: float = 1,
              fill_style=wx.BRUSHSTYLE_SOLID, border_style=wx.PENSTYLE_SOLID, corner_radius: float = 0):
    """Draw a rectangle with the given graphics context.

    Either fill or border must be specified to avoid drawing an entirely transparent rectangle.

    Args:
        gc: The graphics context.
        rect: The rectangle to draw.
        fill: If specified, the fill color of the rectangle.
        border: If specified, the border color of the rectangle.
        border_width: The width of the borders. Defaults to 1. This cannot be 0 when border
            is specified.
        corner_radius: The corner radius of the rounded rectangle. Defaults to 0.
    """
    assert not(fill is None and border is None), \
        "Both 'fill' and 'border' are None, but at least one of them should be provided"

    assert not (border is not None and border_width == 0), \
        "'border_width' cannot be 0 when 'border' is specified"

    x, y = rect.position
    width, height = rect.size

    pen: wx.Pen
    brush: wx.Brush
    # set up brush and pen if applicable
    if fill is not None:
        brush = gc.CreateBrush(wx.Brush(fill, fill_style))
    else:
        brush = wx.TRANSPARENT_BRUSH
    if border is not None:
        pen = gc.CreatePen(wx.GraphicsPenInfo(border).Width(border_width).Style(border_style))
    else:
        pen = wx.TRANSPARENT_PEN

    gc.SetPen(pen)
    gc.SetBrush(brush)

    # draw rect
    gc.DrawRoundedRectangle(x, y, width, height, corner_radius)


"""Classes for the observer-Subject interface. See https://en.wikipedia.org/wiki/Observer_pattern
"""
T = TypeVar('T')


# TODO add SetObserver, which allows delaying callback and combining multiple notify calls.
# e.g. with group_action()
class Observer(abc.ABC, Generic[T]):
    """Observer abstract base class; encapsulates object of type T."""

    def __init__(self, update_callback: Callable[[T], None]):
        self.update = update_callback


class Subject(Generic[T]):
    """Subject abstract base class; encapsulates object of type T."""
    _observers: List[Observer]
    _item: T

    def __init__(self, item):
        self._observers = list()
        self._item = item

    def attach(self, observer: Observer):
        """Attach an observer."""
        self._observers.append(observer)

    def detach(self, observer: Observer):
        """Detach an observer."""
        self._observers.remove(observer)

    def notify(self) -> None:
        """Trigger an update in each Subject."""

        for observer in self._observers:
            observer.update(self._item)


class SetSubject(Subject[Set[T]]):
    """Subject class that encapsulates a set."""

    def __init__(self, *args):
        super().__init__(set(*args))

    def item_copy(self) -> Set:
        """Return a copy of the encapsulated set."""
        return set(self._item)

    def contains(self, val: T) -> bool:
        return val in self._item

    def set_item(self, item: Set):
        """Update the value of the item, notifying observers if the new value differs from the old.
        """
        equal = self._item == item
        self._item = item
        if not equal:
            self.notify()

    def remove(self, el: T):
        """Remove an element from the set, notifying observers if the set changed."""
        equal = el not in self._item
        self._item.remove(el)
        if not equal:
            self.notify()

    def add(self, el: T):
        """Add an element from the set, notifying observers if the set changed."""
        equal = el in self._item
        self._item.add(el)
        if not equal:
            self.notify()

    def union(self, other: Set[T]):
        prev_len = len(self._item)
        self._item |= other
        if len(self._item) != prev_len:
            self.notify()

    def intersect(self, other: Set[T]):
        prev_len = len(self._item)
        self._item &= other
        if len(self._item) != prev_len:
            self.notify()
    
    def __len__(self):
        return len(self._item)

    def __contains__(self, val: T):
        return val in self._item


# the higher the value, the closer the src handle is to the centroid. 1/2 for halfway in-between
# update also for prd handle
CENTER_RATIO = 2/3 
DUPLICATE_RATIO = 3/4
DUPLICATE_ROT = -math.pi/3

def default_handle_positions(centroid: Vec2, reactants: List[Node], products: List[Node]):
    src_handle_pos = reactants[0].rect.center_point * (1 - CENTER_RATIO) + centroid * CENTER_RATIO
    handle_positions = [(n.rect.center_point + centroid) / 2 for n in reactants]
    react_indices = [n.index for n in reactants]
    for prod in products:
        p_rect = prod.rect
        if prod.index in react_indices:
            # If also a reactant, shift the handle to not have the curves completely overlap
            diff = centroid - p_rect.center_point
            length = diff.norm * DUPLICATE_RATIO
            new_dir = rotate_unit(diff, DUPLICATE_ROT)
            handle_positions.append(p_rect.center_point + new_dir * length)
        else:
            #handle_positions.append((p_rect.center_point + centroid) / 2)
            prd_handle_pos = p_rect.center_point*(1-CENTER_RATIO) + centroid*CENTER_RATIO
            handle_positions.append(prd_handle_pos)

    return [src_handle_pos] + handle_positions



"""Classes for managing plugins for Canvas.

Note: This file imports stuff from rkviewer.plugin.classes, but usually rkviewer.plugin imports from
rkviewer. Beware of circular dependency.
"""
# pylint: disable=maybe-no-member
from collections import defaultdict
import importlib.abc
import importlib.util
import inspect
import os
import sys
from typing import Any, Callable, Dict, List, Optional, Tuple, cast
from marshmallow.schema import Schema
import logging
import inspect
import traceback
import requests

# pylint: disable=no-name-in-module
import wx
from rkviewer.plugin.classes import CommandPlugin, Plugin, PluginCategory, PluginType, WindowedPlugin
# pylint: disable=no-name-in-module
from wx.html import HtmlCell, HtmlWidgetCell, HtmlWindow

from rkviewer.config import add_plugin_schema
from rkviewer.events import (CanvasEvent, DidAddCompartmentEvent,
                             DidAddNodeEvent, DidAddReactionEvent, DidChangeCompartmentOfNodesEvent,
                             DidCommitDragEvent, DidDeleteEvent,
                             DidModifyCompartmentsEvent, DidModifyNodesEvent,
                             DidModifyReactionEvent, DidMoveBezierHandleEvent,
                             DidMoveNodesEvent, DidPaintCanvasEvent,
                             DidRedoEvent, DidResizeCompartmentsEvent,
                             DidResizeNodesEvent, DidUndoEvent,
                             SelectionDidUpdateEvent, bind_handler)
from rkviewer.mvc import IController

import json

class PluginManager:
    plugins: List[Plugin]
    callbacks: Dict[Plugin, Callable[[], None]]

    def __init__(self, parent_window: wx.Window, controller: IController):
        self.plugins = list()
        self.callbacks = dict()
        self.parent_window = parent_window
        self.controller = controller
        bind_handler(DidAddNodeEvent, self.make_notify('on_did_add_node'))
        bind_handler(DidMoveNodesEvent, self.make_notify('on_did_move_nodes'))
        bind_handler(DidResizeNodesEvent, self.make_notify('on_did_resize_nodes'))
        bind_handler(DidAddCompartmentEvent, self.make_notify('on_did_add_compartment'))
        bind_handler(DidResizeCompartmentsEvent, self.make_notify('on_did_resize_compartments'))
        bind_handler(DidAddReactionEvent, self.make_notify('on_did_add_reaction'))
        bind_handler(DidUndoEvent, self.make_notify('on_did_undo'))
        bind_handler(DidRedoEvent, self.make_notify('on_did_redo'))
        bind_handler(DidDeleteEvent, self.make_notify('on_did_delete'))
        bind_handler(DidCommitDragEvent, self.make_notify('on_did_commit_drag'))
        bind_handler(DidPaintCanvasEvent, self.make_notify('on_did_paint_canvas'))
        bind_handler(SelectionDidUpdateEvent, self.make_notify('on_selection_did_change'))
        bind_handler(DidMoveBezierHandleEvent, self.make_notify('on_did_move_bezier_handle'))
        bind_handler(DidModifyNodesEvent, self.make_notify('on_did_modify_nodes'))
        bind_handler(DidModifyReactionEvent, self.make_notify('on_did_modify_reactions'))
        bind_handler(DidModifyCompartmentsEvent, self.make_notify('on_did_modify_compartments'))
        bind_handler(DidChangeCompartmentOfNodesEvent,
                     self.make_notify('on_did_change_compartment_of_nodes'))
        self.logger = logging.getLogger('plugin-manager')
        self.error_callback = lambda _: None  # By default don't do anything

    def bind_error_callback(self, callback):
        """Bind a dialog callback for when there is an error.

        If an error occurs before such a callback is bound, it is only logged.
        """
        self.error_callback = callback

    # Also TODO might want a more sophisticated file system structure, including data storage and
    # temp folder
    def load_from(self, load_dir: str) -> bool:
        """Load plugins from the given directory. Returns False if the dir does not exist.
        """
        dirname = os.path.dirname(__file__)
        if os.name == 'nt': #windows
            dir_path = os.path.join(dirname, '..\\', load_dir)
        else:
            dir_path = os.path.join(dirname, '../', load_dir)

        self.plugin_dir = dir_path
        if not os.path.exists(dir_path):
            return False

        plugin_classes = list()
        for f in os.listdir(dir_path):
            if not f.endswith('.py'):
                continue
            mod_name = '_rkviewer.plugin_{}'.format(f[:-2])  # remove extension
            spec = importlib.util.spec_from_file_location(mod_name, os.path.join(dir_path, f))
            assert spec is not None
            mod = importlib.util.module_from_spec(spec)
            assert spec.loader is not None
            loader = cast(importlib.abc.Loader, spec.loader)
            try:
                loader.exec_module(mod)
            except Exception as e:
                except_str = ''.join(traceback.format_exception(None, e, e.__traceback__))
                errmsg = "Failed to load plugin '{}':\n{}".format(f, except_str)
                self.logger.error(errmsg)
                self.error_callback(errmsg)
                continue

            def pred(o): return o.__module__ == mod_name and issubclass(o, Plugin)
            def wrap_exception(pname, method):
                def ret(*args, **kwargs):
                    try:
                        return method(*args, **kwargs)
                    except Exception as e:
                        errmsg = ''.join(traceback.format_exception(None, e, e.__traceback__))
                        errmsg = "Caught error in plugin '{}':\n".format(pname) + errmsg
                        self.logger.error(errmsg)
                        self.error_callback(errmsg)
                return ret

            cur_classes = [m[1] for m in inspect.getmembers(mod, inspect.isclass) if pred(m[1])]
            for cls in cur_classes:
                if inspect.isabstract(cls):
                    logging.warning("Plugin in file '{}' is an abstract class. Did not load.".format(f))
                    continue

                if not hasattr(cls, 'metadata'):
                    logging.warning("Plugin in file '{}' does not have a `metadata` class attribute. "
                        "Did not load. See plugin documentation for more information.".format(f))
                    continue

                for method_name, method in inspect.getmembers(cls, inspect.isroutine):
                    setattr(cls, method_name, wrap_exception(cls.metadata.name, method))

                plugin_classes.append(cls)

        # ...

        logging.getLogger('plugin').info("Found {} valid plugins in '{}'. Loading plugins...".format(
            len(plugin_classes), dir_path))

        #self.plugin_dir = dir_path

        self.plugins = list()
        for cls in plugin_classes:
            try:
                plugin = cls()
                if not hasattr(plugin, 'ptype'):
                    logging.warning("Plugin '{}' has no `ptype` attribute. Did not load. "
                        "Did you forget to call `super().__init__()`?".format(cls.metadata.name))
                    continue
                self.plugins.append(cls())
            except Exception as e:
                self.logger.error('Error when creating plugin object: {}'.format(e))

        # Duplicate names
        if len(set(p.metadata.name for p in self.plugins)) < len(self.plugins):
            pass  # TODO fail when there is duplicate name.

        # Create and register callbacks
        for plugin in self.plugins:
            callback: Callable[[], None]
            if plugin.ptype == PluginType.COMMAND:
                plugin = cast(CommandPlugin, plugin)
                callback = self.make_command_callback(plugin)
            else:
                plugin = cast(WindowedPlugin, plugin)
                callback = self.make_windowed_callback(plugin, self.parent_window)
            self.callbacks[plugin] = callback

        # load schema
        for plugin in self.plugins:
            schema = plugin.get_settings_schema()
            if isinstance(schema, Schema):
                add_plugin_schema(plugin.metadata.name, schema)
        return True

    def make_notify(self, handler_name: str):
        """Make event notification function for plugin.

        handler_name should be the name of a method defined by Plugin. This would then create a
        callback function that goes over each plugin and call that function. This callback should
        be bound to its associated event.

        The handler function is called with the event itself as argument.
        """
        assert callable(getattr(Plugin, handler_name, None)), "{} is not a method defined by \
Plugin!".format(handler_name)

        def ret(evt: CanvasEvent):
            for plugin in self.plugins:
                getattr(plugin, handler_name)(evt)

        return ret

    def register_menu(self, menu: wx.Menu):
        def _get_callback(plugin):
            return lambda _: self.callbacks[plugin]()

        sorted_plugins = sorted(self.plugins, key=lambda p: p.metadata.name)
        for plugin in sorted_plugins:
            item = menu.Append(wx.ID_ANY, plugin.metadata.name)
            menu.Bind(wx.EVT_MENU, _get_callback(plugin), item)

    def make_command_callback(self, command: CommandPlugin) -> Callable[[], None]:
        def command_cb():
            with self.controller.group_action():
                command.run()

        return command_cb

    def make_windowed_callback(self, windowed: WindowedPlugin,
                               parent: wx.Window) -> Callable[[], None]:
        title = windowed.metadata.name
        dialog_exists = False
        dialog: Optional[wx.Window] = None

        def windowed_cb():
            nonlocal dialog_exists, dialog

            if not dialog_exists:
                dialog = wx.Dialog(parent, title=title)
                dialog_exists = True
                window = windowed.create_window(dialog)
                if window is None or not isinstance(window, wx.Window):
                    raise ValueError('create_window() of plugin {} did not return wx.Window '
                                     'type!'.format(windowed.metadata.name))
                windowed.dialog = dialog  # Set the related dialog

                sizer = wx.BoxSizer(wx.VERTICAL)
                sizer.Add(window)
                dialog.SetSize(window.GetSize())
                dialog.SetSizer(sizer)
                dialog.Centre()

                windowed.on_did_create_dialog()
                dialog.Show()

                def close_cb(e):
                    nonlocal dialog_exists
                    windowed.on_will_close_window(e)
                    dialog_exists = False
                dialog.Bind(wx.EVT_CLOSE, close_cb)
            else:
                assert dialog is not None
                dialog.SetFocus()
        return windowed_cb

    def create_dialog(self, parent):
        return PluginDialog(parent, self.plugins)

    def create_install_dialog(self, parent):
        return PluginInstallationDialog(parent, self.plugins, self.plugin_dir)

    def get_plugins_by_category(self) -> Dict[PluginCategory, List[Tuple[str, Callable[[], None], Optional[wx.Bitmap]]]]:
        """Returns a dictionary that maps each category to the list of plugins.

        Each plugin in the lists is a tuple (short_name, bitmap, callback). The
        """
        ret = defaultdict(list)
        for plugin in self.plugins:
            sname = plugin.metadata.short_name if plugin.metadata.short_name else plugin.metadata.name
            assert isinstance(plugin.metadata.category, PluginCategory)
            ret[plugin.metadata.category].append(
                (sname, self.callbacks[plugin], plugin.metadata.icon))
        return ret


class PluginDialog(wx.Dialog):
    def __init__(self, parent, plugins: List[Plugin]):
        super().__init__(parent, title='Manage Plugins', size=(900, 550))
        notebook = wx.Listbook(self, style=wx.LB_LEFT)
        notebook.GetListView().SetFont(wx.Font(wx.FontInfo(10)))
        try:
            notebook.GetListView().SetColumnWidth(0, 125)
        except:
            pass

        sizer = wx.BoxSizer()
        sizer.Add(notebook, proportion=1, flag=wx.EXPAND)

        for plugin in plugins:
            page = PluginPage(notebook, plugin)
            notebook.AddPage(page, text=plugin.metadata.name)

        self.SetSizer(sizer)


class PluginPage(HtmlWindow):
    def __init__(self, parent: wx.Window, plugin: Plugin):
        super().__init__(parent)

        html = '''
        <h3>{name}</h3>
        <div>{author}｜v{version}</div>
        <hr/>
        <div>
            {description}
        </div>
        '''.format(
            name=plugin.metadata.name,
            version=plugin.metadata.version,
            author=plugin.metadata.author,
            description=plugin.metadata.long_desc,
        )

        self.SetPage(html)
        # inherit parent background color for better look
        self.SetBackgroundColour(parent.GetBackgroundColour())

class PluginInstallationDialog(wx.Dialog):
    def __init__(self, parent, plugins: List[Plugin], plugin_dir: str):
        super().__init__(parent, title='Add Plugins', size=(900, 550))
        notebook = wx.Listbook(self, style=wx.LB_LEFT)
        notebook.GetListView().SetFont(wx.Font(wx.FontInfo(10)))
        try:
            notebook.GetListView().SetColumnWidth(0, 125)
        except:
            pass

        sizer = wx.BoxSizer()
        sizer.Add(notebook, proportion=1, flag=wx.EXPAND)

        downloadpage = DownloadPluginPage(notebook, installed=plugins, plugin_dir=plugin_dir)
        notebook.AddPage(downloadpage, text="Find New Plugins")

        add_from_local = AddLocalPage(notebook, plugin_dir=plugin_dir)
        notebook.AddPage(add_from_local, text="Install From Local File")

        self.SetSizer(sizer)

class AddLocalPage(HtmlWindow):
    def __init__(self, parent: wx.Window, plugin_dir: str):
        super().__init__(parent)
        self.plugin_dir = plugin_dir
        self.parent = parent
        self.selected = []
        html = '<h2>Add A Plugin</h2><h4><u>Browse Files</u></h4>'
        self.SetPage(html)

    def OnCellClicked(self, cell, x, y, event):
        with wx.FileDialog(self.parent, "Choose File", wildcard="*.py",
                style=wx.FD_OPEN | wx.FD_FILE_MUST_EXIST) as fileDialog:
            if fileDialog.ShowModal() == wx.ID_CANCEL:
                return False
            # Proceed loading the file chosen by the user
            pathname = fileDialog.GetPath()
            filename = fileDialog.GetFilename()
            try:
                with open(pathname, 'r') as file:
                    txt = file.read()
                savepth = os.path.join(self.plugin_dir, filename)
                with open(savepth, 'w') as file_dest:
                    file_dest.write(txt)
                    file_dest.close()
            except IOError:
                wx.LogError("Cannot open file.")
        self.selected.append(filename)
        self.UpdateDisplay()
        return True

    def UpdateDisplay(self):
        html = '<h2>Add A Plugin</h2><h4><u>Browse Files</u></h4><b>Selected:</b>'
        for item in self.selected:
            html += '<p>{}</p>'.format(item)
        html += '<p><b>Close this window to import selected files</b></p>'
        self.SetPage(html)

class DownloadPluginPage(HtmlWindow):
    def __init__(self, parent: wx.Window, installed: List[Plugin], plugin_dir: str):
        super().__init__(parent)
        self.plugin_dir = plugin_dir
        html ='<h2>{title}</h2>'.format(title="Browse Plugins")
        self.SetPage(html)
        installed_names = dict()
        for p in installed:
            installed_names.update({p.metadata.name: p})
        try:
            # for just displaying info
            r = requests.get('https://raw.githubusercontent.com/sys-bio/SBcoyote-plugins/main/metadata.json')
            metadata = r.json()
            for m in metadata:
                item = '''
                        <div>
                        <h3>{name}</h3>
                        <br>{author}｜v{version}
                        <p>{long_desc}</p>
                        '''.format(name = metadata[m]['name'],
                                            author = metadata[m]['author'],
                                            version = metadata[m]['version'],
                                            long_desc = metadata[m]['long_desc'])
                if metadata[m]['name'] in installed_names:
                    prev_install = installed_names[metadata[m]['name']]
                    if not prev_install.metadata.version == metadata[m]['version']:
                        item += '''
                                <div align="right">v{v} Installed<br>
                                <a href={file}>Update</a></div>
                                '''.format(v = prev_install.metadata.version,
                                            file = m)
                    else:
                        item += '''
                                <div align="right">v{v} Installed</div>
                                '''.format(v = prev_install.metadata.version)
                else:
                    item += '<div align="right"><a href={file}>Install</a></div>'.format(file = m)
                item += '</div><hr>'
                self.AppendToPage(item)
        except: # TODO define error, checking if there is internet connection
            html = '<h2>Browse Plugins</h2><p>Could not connect to server, please try again later!</p>'
            self.SetPage(html)
        self.SetBackgroundColour(parent.GetBackgroundColour())

    def OnLinkClicked(self, link):
        filename = link.GetHref()
        try:
            p = requests.get("https://raw.githubusercontent.com/sys-bio/SBcoyote-plugins/main/all-plugins/{}".format(filename))
        except:
            wx.MessageBox("Could not connect to server. Please try again later.")
        try:
            # path = os.path.join(self.plugin_dir, filename)
            path = os.path.join(self.plugin_dir, filename)
            with open(path, "w") as f:
                f.write(p.text)
                f.close()
            wx.MessageBox("Done. Successfully installed {}.".format(filename), "Complete", wx.OK | wx.ICON_INFORMATION)
        except:
            wx.MessageBox("The program encountered an error. Please try again later.")
